__all__ = [
    "find_module_state_dict",
    "load_checkpoint_to_module",
    "load_model_state_dict",
    "load_state_dict_to_module",
    "show_state_dict_info",
    "state_dicts_are_equal",
]

import logging
from collections.abc import Sequence
from pathlib import Path
from typing import Optional, Union

import torch
from coola import objects_are_equal
from tabulate import tabulate
from torch.nn import Module

from gravitorch import constants as ct
from gravitorch.nn.utils.module_helpers import get_module_device
from gravitorch.utils.mapping import remove_keys_starting_with
from gravitorch.utils.path import sanitize_path

logger = logging.getLogger(__name__)


def find_module_state_dict(state_dict: Union[dict, list, tuple], module_keys: set) -> dict:
    r"""Tries to find automatically the part of the state dict related to a
    module.

    The user should specify the set of module's keys:
    ``set(module.state_dict().keys())``. This function assumes that
    the set of keys only exists at one location in the state dict.
    If the set of keys exists at several locations in the state dict,
    only the first one is returned.

    Args:
        state_dict (dict, list or tuple): Specifies the state dict.
            This function is called recursively on this input to find
            the queried state dict.
        module_keys (set): Specifies the set of module keys.

    Returns:
        dict: The part of the state dict related to a module if it is
            found, otherwise an empty dict.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import find_module_state_dict
        >>> state = {
        ...     'model': {
        ...         'network': {
        ...             'weight': torch.ones(5, 4),
        ...             'bias': 2 * torch.ones(5),
        ...         }
        ...     }
        ... }
        >>> module = torch.nn.Linear(4, 5)
        >>> find_module_state_dict(state, module_keys=set(module.state_dict().keys()))
        {'weight': torch.ones(5, 4), 'bias': 2 * torch.ones(5)}
    """
    if isinstance(state_dict, dict):
        if set(state_dict.keys()) == module_keys:
            return state_dict
        for value in state_dict.values():
            state_dict = find_module_state_dict(value, module_keys)
            if state_dict:
                return state_dict
    elif isinstance(state_dict, (list, tuple, set)):
        for value in state_dict:
            state_dict = find_module_state_dict(value, module_keys)
            if state_dict:
                return state_dict
    return {}


def load_checkpoint_to_module(
    path: Union[Path, str],
    module: Module,
    strict: bool = True,
    key: Optional[Union[str, Sequence[str]]] = None,
) -> None:
    r"""Loads the weights store in a checkpoint file into the given module.

    The checkpoint should contain the state dict of the module, which
    is the recommended way to store the model state/weights
    https://pytorch.org/docs/stable/notes/serialization.html#serialization-semantics.
    This function raises an error (``RuntimeError``) if the weights
    are not compatible with the module. This function assumes the
    model is on a single device and may not work in a multi-device
    setting. You will have to create your own function to manage the
    multi-device setting. This function automatically moves the
    weights to the module device if they are on another device.

    Args:
        module (``torch.nn.Module``): Specifies the module. This
            function changes the weights of this module.
        path (str): Specifies the path to the checkpoint with the
            module weights. The checkpoint should be a PyTorch file.
        strict (bool, optional): whether to strictly enforce that the
            keys in ``state_dict`` match the keys returned by this
            module's :meth:`~torch.nn.Module.state_dict` function.
            Default: ``True``
        key (str or list or tuple or ``None``, optional): Specifies
            the key of the state dict to load. The state dict can
            contain data that are not about the module, so they may
            need to be excluded. For nested case, it is possible to
            specify the list of keys to get the good part of the state
            dict. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import load_checkpoint_to_module
        >>> net = torch.nn.Linear(4, 5)
        >>> load_checkpoint_to_module(path="/path/to/checkpoint.pt", module=net)
        # To load only the value defined at state_dict["model"]
        >>> load_checkpoint_to_module(path="/path/to/checkpoint.pt", module=net, key="model")
        # To load only the value defined at state_dict["model"]["network"]
        >>> load_checkpoint_to_module(
        ...     path="/path/to/checkpoint.pt",
        ...     module=net,
        ...     key=["model", "network"],
        ... )
        >>> load_checkpoint_to_module(
        ...     path="/path/to/checkpoint.pt",
        ...     module=net,
        ...     key=("model", "network"),
        ... )
    """
    # Get the device of the module. It assumes the module is on a single device.
    device = get_module_device(module)
    # Load the state dict.
    path = sanitize_path(path)
    logger.info(f"Loading checkpoint from {path}...")
    state_dict = torch.load(path, map_location=device)

    if key:  # Go to a given key of the state dict
        key = (key,) if isinstance(key, str) else key
        for k in key:
            state_dict = state_dict[k]

    load_state_dict_to_module(state_dict=state_dict, module=module, strict=strict)


def load_state_dict_to_module(state_dict: dict, module: Module, strict: bool = True) -> None:
    r"""Loads a state dict into a given module.

    This function will automatically try to find the module state dict
    in the given state dict.

    Args:
        state_dict (dict): Specifies the state dict with the module
            state dict.
        module (``torch.nn.Module``): Specifies the module. This
            function changes the weights of this module.
        strict (bool, optional): whether to strictly enforce that the
            keys in ``state_dict`` match the keys returned by this
            module's :meth:`~torch.nn.Module.state_dict` function.
            Default: ``True``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.utils import load_state_dict_to_module
        >>> module = torch.nn.Linear(4, 5)
        >>> state_dict = {
        ...     "weight": torch.tensor(
        ...         [
        ...             [-0.0328, 0.1781, -0.1038, 0.4252],
        ...             [0.0242, -0.2810, 0.3932, 0.2912],
        ...             [-0.0474, -0.0365, -0.0279, -0.1859],
        ...             [0.2816, 0.2591, -0.3863, 0.4778],
        ...             [0.3242, 0.0191, 0.0883, 0.0032],
        ...         ]
        ...     ),
        ...     "bias": torch.tensor([0.4347, -0.3001, -0.3297, 0.2118, -0.3951]),
        ... }
        >>> load_state_dict_to_module(path="/path/to/checkpoint.pt", module=module)
    """
    try:
        module.load_state_dict(state_dict, strict)
    except RuntimeError:
        logger.warning(
            "Could not load the state dict. Try to find automatically the part of the state dict "
            "that matches with the module."
        )
        state_dict = find_module_state_dict(state_dict, set(module.state_dict().keys()))
        if not state_dict:
            logger.info("Could not find a part of the state dict that matches with the module.")
            raise
        logger.info(
            "Found a part of the state dict that matches with the module. Try to load it in the "
            "module."
        )
        module.load_state_dict(state_dict, strict)
    logger.info("The weights are loaded in the module.")


def state_dicts_are_equal(module1: Module, module2: Module) -> bool:
    r"""Indicates if the state dict of 2 modules are equal or not.

    Args:
        module1 (``torch.nn.Module``): Specifies the first module to
            compare.
        module2 (``torch.nn.Module``): Specifies the second module to
            compare.

    Returns:
        bool: ``True`` if the state dict of 2 modules are equal,
            otherwise ``False``
    """
    return objects_are_equal(module1.state_dict(), module2.state_dict())


def show_state_dict_info(state_dict: dict, tablefmt: str = "simple") -> None:
    r"""Shows information about the state dict.

    Args:
        state_dict (dict): Specifies the state dict to analyze.
        tablefmt (str, optional): Specifies the table format.
            Default: ``'simple'``
    """
    stats = [["key", "shape", "dtype"]]
    for key, value in state_dict.items():
        if torch.is_tensor(value):
            stats.append([key, list(value.shape), value.dtype])
    logger.info(f'State dict info\n{tabulate(stats, headers="firstrow", tablefmt=tablefmt)}\n')


def load_model_state_dict(
    path: Union[Path, str],
    module: Module,
    exclude_key_prefixes: Optional[Sequence[str]] = None,
    strict: bool = True,
) -> None:
    r"""Loads a model state dict.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the
            checkpoint with the module weights. The checkpoint
            should be a PyTorch file.
        module (``torch.nn.Module``): Specifies the module. This
            function changes the weights of this module.
        exclude_key_prefixes (sequence or ``None``, optional):
            Specifies the list of key prefixes to exclude when
            loading the state dict. Default: ``None``
        strict (bool, optional): whether to strictly enforce that
            the keys in ``state_dict`` match the keys returned by
            this module's :meth:`~torch.nn.Module.state_dict`
            function. Default: ``True``
    """
    # Load the state dict.
    path = sanitize_path(path)
    logger.info(f"Loading checkpoint from {path}...")
    # Get the device of the module. It assumes the module is on a single device.
    device = get_module_device(module)
    state_dict = torch.load(path, map_location=device)

    # Prepare the state dict
    state_dict = state_dict["modules"][ct.MODEL]  # TODO: make more generic
    exclude_key_prefixes = exclude_key_prefixes or []
    for prefix in exclude_key_prefixes:
        state_dict = remove_keys_starting_with(state_dict, prefix)

    show_state_dict_info(state_dict)
    module.load_state_dict(state_dict, strict)



from __future__ import annotations

__all__ = ["BaseModelCreator", "setup_model_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from torch import nn

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class BaseModelCreator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create a model.

    Note that it is not the unique approach to create a model. Feel free
    to use other approaches if this approach does not fit your needs.
    """

    @abstractmethod
    def create(self, engine: BaseEngine) -> nn.Module:
        r"""Creates a model on the device(s) where it should run.

        This method is responsible to register the event handlers
        associated to the model. This method is also responsible to
        move the model parameters to the device(s).

        Args:
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.

        Returns:
            ``torch.nn.Module``: The created model.
        """


def setup_model_creator(creator: BaseModelCreator | dict) -> BaseModelCreator:
    r"""Sets up the model creator.

    The model creator is instantiated from its configuration by using
    the ``BaseModelCreator`` factory function.

    Args:
        creator (``BaseModelCreator`` or dict): Specifies the model
            creator or its configuration.

    Returns:
        ``BaseModelCreator``: The instantiated model creator.
    """
    if isinstance(creator, dict):
        logger.info(
            f"Initializing a model creator from its configuration... {str_target_object(creator)}"
        )
        creator = BaseModelCreator.factory(**creator)
    return creator



r"""This module implements a simple training loop."""

__all__ = ["VanillaTrainingLoop"]

import logging
import sys
from collections.abc import Callable, Iterable
from typing import Any, Optional, Union

import torch
from torch.nn import Module
from torch.optim import Optimizer
from tqdm import tqdm

from gravitorch import constants as ct
from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.utils.device_placement import (
    AutoDevicePlacement,
    BaseDevicePlacement,
    setup_device_placement,
)
from gravitorch.utils.loop_observers import BaseLoopObserver
from gravitorch.utils.profilers import BaseProfiler
from gravitorch.utils.training_loops.base import BaseBasicTrainingLoop

logger = logging.getLogger(__name__)


class VanillaTrainingLoop(BaseBasicTrainingLoop):
    r"""Implements a simple training loop to train a model on a dataset.

    Args:
        set_grad_to_none (bool, optional): If ``True``, set the
            gradients to ``None``, otherwise set the gradients to
            zero. Setting the gradients to ``None`` will in general
            have lower memory footprint, and can modestly improve
            performance. Default: ``False``
        batch_device_placement (``BaseDevicePlacement`` or dict or
            ``None``, optional): Specifies the batch device placement
            module. This module moves the batch on a target device.
            The target device should be compatible with the model.
            If ``None``, an ``AutoDevicePlacement`` object is
            instantiated. Default: ``None``
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"train"``
        clip_grad (dict or None, optional): Specifies the
            configuration to clip the gradient. If ``None``, no
            gradient clipping is used during the training.
            Default: ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional): Specifies
            the profiler or its configuration. If ``None``, the
            ``NoOpProfiler`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        set_grad_to_none: bool = False,
        batch_device_placement: Union[BaseDevicePlacement, dict, None] = None,
        tag: str = "train",
        clip_grad: Optional[dict] = None,
        observer: Union[BaseLoopObserver, dict, None] = None,
        profiler: Union[BaseProfiler, dict, None] = None,
    ):
        super().__init__(tag=tag, clip_grad=clip_grad, observer=observer, profiler=profiler)
        self._set_grad_to_none = bool(set_grad_to_none)
        self._batch_device_placement = setup_device_placement(
            batch_device_placement or AutoDevicePlacement()
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  tag={self._tag},\n"
            f"  set_grad_to_none={self._set_grad_to_none},\n"
            f"  batch_device_placement={self._batch_device_placement},\n"
            f"  clip_grad_fn={self._clip_grad_fn},\n"
            f"  clip_grad_args={self._clip_grad_args},\n"
            f"  observer={self._observer},\n"
            f"  profiler={self._profiler},\n"
            ")"
        )

    def _prepare_model_optimizer_data_loader(
        self, engine: BaseEngine
    ) -> tuple[Module, Optimizer, Iterable]:
        logger.info("Preparing the model, optimizer, and data loader...")
        data_loader = engine.data_source.get_data_loader(loader_id=self._tag, engine=engine)
        prefix = f"({dist.get_rank()}/{dist.get_world_size()}) " if dist.is_distributed() else ""
        data_loader = tqdm(
            data_loader,
            desc=f"{prefix}Training [{engine.epoch}/{engine.max_epochs}]",
            position=dist.get_rank(),
            file=sys.stdout,
        )
        logger.info("Training data loader has been created")
        return engine.model, engine.optimizer, data_loader

    def _train_one_batch(
        self, engine: BaseEngine, model: Module, optimizer: Optimizer, batch: Any
    ) -> dict:
        engine.fire_event(EngineEvents.TRAIN_ITERATION_STARTED)
        optimizer.zero_grad(self._set_grad_to_none)
        output = model(self._batch_device_placement.send(batch))
        engine.fire_event(EngineEvents.TRAIN_FORWARD_COMPLETED)

        loss = output[ct.LOSS]
        if torch.isnan(loss):
            logger.warning(
                "NaN detected. The gradient is not computed for this batch "
                f"(iteration: {engine.iteration})"
            )
            engine.fire_event(EngineEvents.TRAIN_ITERATION_COMPLETED)
            return output

        loss.backward()
        if self._clip_grad_fn:
            self._clip_grad_fn(model.parameters(), *self._clip_grad_args)
        engine.fire_event(EngineEvents.TRAIN_BACKWARD_COMPLETED)

        optimizer.step()
        engine.fire_event(EngineEvents.TRAIN_ITERATION_COMPLETED)

        return output

    def _setup_clip_grad(self, clip_grad: dict) -> tuple[Optional[Callable], tuple]:
        if not clip_grad:
            return None, ()

        name = clip_grad["name"]
        if name == "clip_grad_value":
            clip_grad_fn = torch.nn.utils.clip_grad_value_
            clip_grad_args = (clip_grad.get("clip_value", 0.25),)
            logger.info(f"clip gradient by value {clip_grad_args[0]}")
            return clip_grad_fn, clip_grad_args
        if name == "clip_grad_norm":
            clip_grad_fn = torch.nn.utils.clip_grad_norm_
            clip_grad_args = clip_grad.get("max_norm", 1), clip_grad.get("norm_type", 2)
            logger.info(
                f"clip gradient by maximum norm {clip_grad_args[0]} "
                f"(norm type: {clip_grad_args[1]})"
            )
            return clip_grad_fn, clip_grad_args
        raise ValueError(
            f"Incorrect clip grad name ({name}). The valid values are ``clip_grad_value`` "
            "and ``clip_grad_norm``"
        )



__all__ = [
    "BaseXavierParameterInitializer",
    "XavierNormalParameterInitializer",
    "XavierUniformParameterInitializer",
    "recursive_xavier_normal_",
    "recursive_xavier_uniform_",
]

import logging

from torch import nn

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.parameter_initializers.base import BaseDefaultParameterInitializer

logger = logging.getLogger(__name__)


class BaseXavierParameterInitializer(BaseDefaultParameterInitializer):
    r"""Implements a model parameter initializer with the Xavier Normal or
    uniform strategy.

    Args:
        gain (float, optional): Specifies the gain or scaling factor.
            Default: ``1``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        show_stats (bool, optional): If ``True``, the parameter
            statistics are shown at the end of the initialization.
            Default: ``True``
    """

    def __init__(self, gain: float = 1.0, learnable_only: bool = True, show_stats: bool = True):
        super().__init__(show_stats=show_stats)
        self._gain = float(gain)
        self._learnable_only = bool(learnable_only)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(gain={self._gain}, "
            f"learnable_only={self._learnable_only})"
        )


class XavierNormalParameterInitializer(BaseXavierParameterInitializer):
    r"""Implements a model parameter initializer with the Xavier Normal
    strategy."""

    def _initialize(self, engine: BaseEngine) -> None:
        logger.info("Initializing model parameters with the Xavier Normal strategy...")
        recursive_xavier_normal_(
            module=engine.model, gain=self._gain, learnable_only=self._learnable_only
        )


class XavierUniformParameterInitializer(BaseXavierParameterInitializer):
    r"""Implements a model parameter initializer with the Xavier uniform
    strategy."""

    def _initialize(self, engine: BaseEngine) -> None:
        logger.info("Initializing model parameters with the Xavier uniform strategy...")
        recursive_xavier_uniform_(
            module=engine.model, gain=self._gain, learnable_only=self._learnable_only
        )


def recursive_xavier_normal_(
    module: nn.Module, gain: float = 1.0, learnable_only: bool = True
) -> None:
    r"""Initialize the parameters of the module with the Xavier Normal
    initialization.

    Args:
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        gain (float, optional): Specifies the gain or scaling factor.
            Default: ``1``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized,
            otherwise all the parameters are initialized.
            Default: ``True``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.parameter_initializers import recursive_xavier_normal_
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> recursive_xavier_normal_(net)
    """
    for params in module.parameters():
        if params.ndim > 1 and (not learnable_only or learnable_only and params.requires_grad):
            nn.init.xavier_normal_(params.data, gain=gain)


def recursive_xavier_uniform_(
    module: nn.Module, gain: float = 1.0, learnable_only: bool = True
) -> None:
    r"""Initialize the parameters of the module with the Xavier uniform
    initialization.

    Args:
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        gain (float, optional): Specifies the gain or scaling factor.
            Default: ``1``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.parameter_initializers import recursive_xavier_uniform_
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> recursive_xavier_uniform_(net)
    """
    for params in module.parameters():
        if params.ndim > 1 and (not learnable_only or learnable_only and params.requires_grad):
            nn.init.xavier_uniform_(params.data, gain=gain)



__all__ = ["app_vae_time_loss"]

import torch

from gravitorch.nn.functional.loss_helpers import basic_loss_reduction


def app_vae_time_loss(
    lmbda: torch.Tensor,
    target: torch.Tensor,
    log_input: bool = False,
    delta: float = 0.1,
    eps: float = 1e-8,
    max_log_value: float = 20.0,
    reduction: str = "mean",
) -> torch.Tensor:
    r"""Computes the APP-VAE time loss value.

    Args:
        lmbda (``torch.Tensor`` of type float): Specifies the
            predicted lambda i.e. the rate of the Exponential
            distribution.
        target (``torch.Tensor`` of type float and same shape as
            ``lmbda``): Specifies the target values.
        log_input (bool, optional):  If ``True``, the expected input
            is ``log(lambda)``, otherwise it is ``lambda``.
            Default: ``False``
        delta (float, optional): Specifies a small time interval used
            to compute the NLL. Default: ``0.1``
        eps (float, optional): Small value to avoid evaluation of
            :math:`\log(0)`. Default: ``1e-8``
        max_log_value (float, optional): Specifies the maximum value
            used to clip ``log_rate`` before to compute the
            exponential when :attr:`log_input = True`.
            Default: ``20.0``
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``

    Returns:
        ``torch.Tensor`` of type float: The negative log-likelihood
            with Exponential distribution of target. The shape of
            the tensor depends on the reduction strategy.
    """
    if delta <= 0:
        raise ValueError(f"delta has to be greater than 0 but received: {delta}")
    if log_input:
        lmbda = lmbda.clamp(max=max_log_value).exp()
    nll = lmbda.mul(target) - (1 - lmbda.mul(-delta).exp() + eps).log()
    return basic_loss_reduction(nll, reduction)



__all__ = ["DirFilterIterDataPipe", "FileFilterIterDataPipe", "PathListerIterDataPipe"]

from collections.abc import Iterator
from pathlib import Path

from torch.utils.data import IterDataPipe

from gravitorch.utils.format import str_add_indent


class DirFilterIterDataPipe(IterDataPipe[Path]):
    r"""Implements an ``IterDataPipe`` to keep only the directory.

    Args:
        datapipe (``IterDataPipe``): Specifies the source
            ``IterDataPipe``.
    """

    def __init__(self, datapipe: IterDataPipe[Path]):
        self._datapipe = datapipe

    def __iter__(self) -> Iterator[Path]:
        for path in self._datapipe:
            if path.is_dir():
                yield path

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n" f"  datapipe={str_add_indent(self._datapipe)},\n)"
        )


class FileFilterIterDataPipe(IterDataPipe[Path]):
    r"""Implements an ``IterDataPipe`` to keep only the files.

    Args:
        datapipe (``IterDataPipe``): Specifies the source
            ``IterDataPipe``.
    """

    def __init__(self, datapipe: IterDataPipe[Path]):
        self._datapipe = datapipe

    def __iter__(self) -> Iterator[Path]:
        for path in self._datapipe:
            if path.is_file():
                yield path

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n" f"  datapipe={str_add_indent(self._datapipe)},\n)"
        )


class PathListerIterDataPipe(IterDataPipe[Path]):
    r"""Implements an ``IterDataPipe`` to list the paths.

    Args:
        datapipe (``IterDataPipe``): Specifies the source
            ``IterDataPipe`` with the root paths.
        pattern (str, optional): Specifies a glob pattern, to return
            only the matching paths. Default: ``'*'``
        deterministic (bool, optional): If ``True``, the paths are
            returned in a deterministic order. Default: ``True``
    """

    def __init__(
        self,
        datapipe: IterDataPipe[Path],
        pattern: str = "*",
        deterministic: bool = True,
    ):
        self._datapipe = datapipe
        self._pattern = pattern
        self._deterministic = bool(deterministic)

    def __iter__(self) -> Iterator[Path]:
        for path in self._datapipe:
            paths = path.glob(self._pattern)
            if self._deterministic:
                paths = sorted(paths)
            yield from paths

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  pattern={self._pattern},\n"
            f"  deterministic={self._deterministic},\n"
            f"  datapipe={str_add_indent(self._datapipe)},\n)"
        )



__all__ = ["BaseProfiler"]

from abc import ABC, abstractmethod

from objectory import AbstractFactory


class BaseProfiler(ABC, metaclass=AbstractFactory):
    r"""Defines the base profiler class.

    Each profiler should be a context manager.

    .. note::

    The profiler is an experimental feature and may change in the
    future without warnings.
    """

    def __enter__(self):
        r"""Starts profiling."""

    def __exit__(self, exc_type, exc_val, exc_tb):
        r"""Ends profiling."""

    @abstractmethod
    def step(self) -> None:
        r"""Signals the profiler that the next profiling step has started."""



__all__ = ["LooperIterDataPipe"]

from collections.abc import Iterator
from typing import TypeVar

from torch.utils.data import IterDataPipe

from gravitorch.utils.format import str_add_indent

T = TypeVar("T")


class LooperIterDataPipe(IterDataPipe[T]):
    r"""Implements a DataPipe that has a fixed length.

    - If the source DataPipe is longer than ``length``, only the first
        ``length`` items of the source DataPipe are used.
    - If the source DataPipe is shorter than ``length``, the items of
        the source DataPipe are repeated until ``length`` items.

    Args:
        datapipe (``torch.utils.data.IterDataPipe``): Specifies
            the source iterable DataPipe.
        length (int): Specifies the length of the DataPipe.

     Example usage:

    .. code-block:: python

        >>> from gravitorch.data.datapipes.iter import Looper, SourceWrapper
        >>> dp = Looper(SourceWrapper([1, 2, 3, 4]), length=10)
        >>> list(dp)
        [1, 2, 3, 4, 1, 2, 3, 4, 1, 2]
        >>> dp = Looper(SourceWrapper([1, 2, 3, 4, 5, 6]), length=4)
        >>> list(dp)
        [1, 2, 3, 4]
    """

    def __init__(self, datapipe: IterDataPipe[T], length: int):
        self._datapipe = datapipe
        if length < 1:
            raise ValueError(
                f"Incorrect length: {length}. The length has to be greater or equal to 1"
            )
        self._length = int(length)

    def __iter__(self) -> Iterator[T]:
        step = 0
        while step < self._length:
            for data in self._datapipe:
                yield data
                step += 1
                if step == self._length:
                    break

    def __len__(self) -> int:
        return self._length

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  length={self._length:,},\n"
            f"  datapipe={str_add_indent(self._datapipe)},\n)"
        )



r"""This module implements some utility functions to do standard element-wise
operations on some data structures."""

__all__ = ["add_objects", "sub_objects"]

from typing import Any


def add_objects(object1: Any, object2: Any) -> Any:
    r"""Performs an element-wise addition of two objects.

    Equivalent to: ``object1 + object2``.

    The current implementation supports the following object types:

        - int
        - float
        - ``torch.Tensor``
        - ``numpy.ndarray``
        - ``list``
        - ``tuple``
        - ``dict``

    The set are not supported because it is not always possible to do
    an element-wise addition. This function assumes that both objects
    have the same structure, and it is possible to do an element-wise
    operation between the values. For example if
    ``object1['my_list']`` is a list, it is expected that
    ``object2['my_list']`` is also a list.

    Args:
        object1: Specifies the first object
        object2: Specifies the second object

    Returns:
        The element-wise addition of the two objects.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.object_operations import add_objects
        >>> add_objects([1, 2, 3], [4, 5, 6])
        [5, 7, 9]
        >>> add_objects(
        ...     (torch.tensor([1, 2]), torch.tensor([3])),
        ...     (torch.tensor([4, 5]), torch.tensor([6])),
        ... )
        (torch.tensor([5, 7]), torch.tensor([9]))
    """
    if isinstance(object1, dict):
        return type(object1)(
            {key: add_objects(value, object2[key]) for key, value in object1.items()}
        )
    if isinstance(object1, (tuple, list)):
        return type(object1)(add_objects(obj1, obj2) for obj1, obj2 in zip(object1, object2))
    return object1 + object2


def sub_objects(object1: Any, object2: Any) -> Any:
    r"""Performs an element-wise substraction of two objects.

    Equivalent to: ``object1 - object2``

    The current implementation supports the following object types:

        - int
        - float
        - ``torch.Tensor``
        - ``numpy.ndarray``
        - ``list``
        - ``tuple``
        - ``dict``

    The set are not supported because it is not always possible to do
    an element-wise addition. This function assumes that both objects
    have the same structure, and it is possible to do an element-wise
    operation between the values. For example if
    ``object1['my_list']`` is a list, it is expected that
    ``object2['my_list']`` is also a list.

    Args:
        object1: Specifies the first object
        object2: Specifies the second object

    Returns:
        The element-wise substraction of the two objects.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.object_operations import sub_objects
        >>> sub_objects([1, 2, 3], [4, 5, 6])
        [-3, -3, -3]
        >>> sub_objects(
        ...     (torch.tensor([4, 5]), torch.tensor([6])),
        ...     (torch.tensor([1, 0]), torch.tensor([3])),
        ... )
        (torch.tensor([3, 5]), torch.tensor([3]))
    """
    if isinstance(object1, dict):
        return type(object1)(
            {key: sub_objects(value, object2[key]) for key, value in object1.items()}
        )
    if isinstance(object1, (tuple, list)):
        return type(object1)(sub_objects(obj1, obj2) for obj1, obj2 in zip(object1, object2))
    return object1 - object2



r"""This module implements a handler to initialize parameters."""

__all__ = ["ParameterInitializer"]

import logging
from typing import Union

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.events import VanillaEventHandler
from gravitorch.utils.format import str_add_indent
from gravitorch.utils.parameter_initializers import (
    BaseParameterInitializer,
    setup_parameter_initializer,
)

logger = logging.getLogger(__name__)


class ParameterInitializer(BaseHandler):
    r"""Implements a handler to initialize the parameters.

    This handler uses a ``BaseParameterInitializer`` object to
    initialize parameters.

    Args:
        parameter_initializer (``BaseParameterInitializer`` or dict):
            Specifies the parameter initializer or its configuration.
        event (str, optional): Specifies the event when to initialize
            the parameters. Default: ``'train_started'``
    """

    def __init__(
        self,
        parameter_initializer: Union[BaseParameterInitializer, dict],
        event: str = EngineEvents.TRAIN_STARTED,
    ):
        self._parameter_initializer = setup_parameter_initializer(parameter_initializer)
        self._event = event

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  parameter_initializer={str_add_indent(self._parameter_initializer)},"
            f"  event={self._event},"
            ")"
        )

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(
                self._parameter_initializer.initialize,
                handler_kwargs={"engine": engine},
            ),
        )



r"""This module defines the base class of the history tracker."""

__all__ = [
    "BaseHistory",
    "EmptyHistoryError",
    "HistoryAllCloseOperator",
    "HistoryEqualityOperator",
    "NotAComparableHistoryError",
]

import logging
from abc import ABC, abstractmethod
from typing import Any, Generic, Optional, TypeVar

from coola import (
    AllCloseTester,
    BaseAllCloseOperator,
    BaseAllCloseTester,
    BaseEqualityOperator,
    BaseEqualityTester,
    EqualityTester,
)
from objectory import OBJECT_TARGET, AbstractFactory
from objectory.utils import full_object_name

T = TypeVar("T")

logger = logging.getLogger(__name__)


class BaseHistory(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Definition of a base class to track a history of values.

    The history tracks the value added as well as the step
    when the value is added. The goal of this class is to track the
    recent history because the loggers (e.g. MLFlow or Tensorboard)
    do not allow to get the last value or the best value. The history
    keeps in memory a recent history of pairs (step, value)
    where step is the index of the step when the value was added. The
    length of the recent history depends on the concrete
    implementation.

    To implement your own history, you will need to define the
    following methods:

        - ``add_value``
        - ``get_last_value``
        - ``get_recent_history``
        - ``is_comparable``
        - ``load_state_dict``
        - ``state_dict``

    If it is a comparable history, you will need to implement
    the following methods too:

        - ``_get_best_value``
        - ``_has_improved``

    You may also need to extend the ``config_dict`` method.

    Args:
        name (str): Specifies the name of the history.
    """

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        r"""str: The name of the history."""
        return self._name

    @abstractmethod
    def add_value(self, value: T, step: Optional[int] = None) -> None:
        r"""Adds a new value to the history.

        Args:
            value: Specifies the value to add to the history.
            step (int or None, optional): Specifies the step value to
                record. Default: ``None``.

        Example usage:

        .. code-block:: python

            >>> import torch
            >>> from gravitorch.utils.history import GenericHistory
            >>> history = GenericHistory('loss')
            >>> history.add_value(value=2)
            >>> history.add_value(value=torch.zeros(2, 3), step=1)
        """

    def clone(self) -> "BaseHistory":
        r"""Clones the current history.

        Returns:
            ``BaseHistory``: A copy of the current history.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import GenericHistory
            >>> history = GenericHistory('epoch')
            >>> history_cloned = history.clone()
        """
        return self.from_dict(self.to_dict())

    @abstractmethod
    def equal(self, other: Any) -> bool:
        r"""Indicates if two histories are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the histories are equal,
                ``False`` otherwise.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import GenericHistory
            >>> history1 = GenericHistory('loss')
            >>> history2 = GenericHistory('accuracy')
            >>> history1.equal(history2)
            False
        """

    def get_best_value(self) -> T:
        r"""Gets the best value of this history.

        It is possible to get the best value only if it is a
        comparable history i.e. it is possible to compare the
        values in the history.

        Returns:
            The best value of this history.

        Raises:
            ``NotAComparableHistory``: if it is not a comparable
                history.
            ``EmptyHistoryError``: if the history is empty

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import MaxScalarHistory
            >>> history = MaxScalarHistory('accuracy')
            >>> history.add_value(value=2, step=0)
            >>> history.add_value(value=4, step=1)
            >>> history.get_best_value()
            4
        """
        if not self.is_comparable():
            raise NotAComparableHistoryError(
                "It is not possible to get the best value because it is not possible to compare "
                f"the values in {self.name} history"
            )
        return self._get_best_value()

    def _get_best_value(self) -> T:
        r"""Gets the best value of this history.

        You need to implement this method for a comparable history.

        Returns:
            The best value of this history.

        Raises:
            ``NotImplementedError``: if this method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def get_last_value(self) -> T:
        r"""Gets the last value.

        Returns:
            The last value added in the history.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import GenericHistory
            >>> history = GenericHistory('loss')
            >>> history.add_value(value=2, step=0)
            >>> history.add_value(value=1.2, step=1)
            >>> history.get_last_value()
            1.2
            >>> history.add_value(value=0.8, step=1)
            >>> history.get_last_value()
            0.8
        """

    @abstractmethod
    def get_recent_history(self) -> tuple[tuple[Optional[int], T], ...]:
        r"""Gets the list of value in the recent history.

        The last value in the tuple is the last value added to the
        history. The length of the recent history depends on the
        concrete implementation.

        Returns:
            tuple: A tuple of the recent history.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import GenericHistory
            >>> history = GenericHistory('loss')
            >>> history.add_value(value=2)
            >>> history.add_value(value=1.2, step=1)
            >>> history.add_value(value=0.8, step=2)
            >>> history.get_recent_history()
            ((None, 2), (1, 1.2), (2, 0.8))
        """

    def has_improved(self) -> bool:
        r"""Indicates if the last value is the best value.

        It is possible to use this method only if it is a comparable
        history i.e. it is possible to compare the values in
        the history.

        Returns:
            bool: ``True`` if the last value is the best value,
                otherwise ``False``.

        Raises:
            ``NotAComparableHistory``: if it is not a comparable
                history.
            ``EmptyHistoryError``: if the history is empty

        .. code-block:: python

            >>> from gravitorch.utils.history import MaxScalarHistory
            >>> history = MaxScalarHistory('accuracy')
            >>> history.add_value(value=2, step=0)
            >>> history.add_value(value=4, step=1)
            >>> history.has_improved()
            True
        """
        if not self.is_comparable():
            raise NotAComparableHistoryError(
                "It is not possible to indicate if the last value is the best value because it "
                f"is not possible to compare the values in {self.name} history"
            )
        return self._has_improved()

    def _has_improved(self) -> bool:
        r"""Indicates if the last value is the best value.

        You need to implement this method for a comparable history.

        Returns:
            bool: ``True`` if the last value is the best value,
                otherwise ``False``.

        Raises:
            ``NotImplementedError``: if this method is not implemented
        """
        raise NotImplementedError

    @abstractmethod
    def is_comparable(self) -> bool:
        r"""Indicates if it is possible to compare the values in the history.

        Note that it is possible to compute the best value only for
        histories that are comparable.

        Returns:
            bool: ``True`` if it is possible to compare the values in
            the history, otherwise ``False``.

        .. code-block:: python

            >>> from gravitorch.utils.history import GenericHistory
            >>> history = GenericHistory('loss')
            >>> history.is_comparable()
            False
        """

    @abstractmethod
    def is_empty(self) -> bool:
        r"""Indicates if the history is empty or not.

        Returns:
            bool: ``True`` if the history is empty, otherwise
                ``False``.

        .. code-block:: python

            >>> from gravitorch.utils.history import GenericHistory
            >>> history = GenericHistory('loss')
            >>> history.is_empty()
            True
        """

    def config_dict(self) -> dict[str, Any]:
        r"""Gets the config of the history.

        The config dictionary should contain all the values necessary
        to instantiate a history with the same parameters
        with the  ``factory`` method. It is expected to contain values
        like the full name of the class and the arguments of the
        constructor. This dictionary should not contain the state
        values. It is possible to get the state values with the
        ``state_dict`` method.

        Returns:
            dict: The config of the history

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import BaseHistory, GenericHistory
            >>> config = GenericHistory('loss').config_dict()
            >>> history = BaseHistory.factory(**config)  # Note that the state is not copied.
        """
        return {
            OBJECT_TARGET: full_object_name(self.__class__),
            "name": self._name,
        }

    @abstractmethod
    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Sets up the history from a dictionary containing the state values.

        Args:
            state_dict (dict): Specifies a dictionary containing state
                keys with values.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import GenericHistory
            >>> history = GenericHistory('loss')
            >>> history.load_state_dict({...})
        """

    @abstractmethod
    def state_dict(self) -> dict[str, Any]:
        r"""Gets a dictionary containing the state values of the history.

        Returns:
            dict: The state values in a dict.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import GenericHistory
            >>> state = GenericHistory('loss').state_dict()
        """

    @classmethod
    def from_dict(cls, data: dict) -> "BaseHistory":
        r"""Instantiates a history from a dictionary.

        Args:
            data (dict): Specifies the dictionary that is used to
                instantiate the history. The dictionary is
                expected to contain the parameters to create
                instantiate the history and the state of the
                history.

        Returns:
            ``BaseHistory``: The instantiated history.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import BaseHistory
            >>> from objectory import OBJECT_TARGET
            >>> history = BaseHistory.from_dict(
            ...     {
            ...         "config": {
            ...             OBJECT_TARGET: "gravitorch.utils.history.generic.GenericHistory",
            ...             "name": "loss",
            ...             "max_size": 7,
            ...         },
            ...         "state": {"history": ((0, 1), (1, 5))},
            ...     }
            ... )
        """
        obj = cls.factory(**data["config"])
        obj.load_state_dict(data["state"])
        return obj

    def to_dict(self) -> dict[str, Any]:
        r"""Exports the current history to a dictionary.

        This method exports all the information to re-create the
        history with the same state. The returned dictionary
        can be used as input of the ``from_dict`` method to resume the
        history.

        Returns:
            dict: A dictionary with the config and the state of the
                history.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.history import BaseHistory, GenericHistory
            >>> history_dict = GenericHistory('loss').to_dict()
            >>> history = BaseHistory.from_dict(**history_dict)
        """
        return {"config": self.config_dict(), "state": self.state_dict()}


class EmptyHistoryError(Exception):
    r"""Generates an error if the history is empty."""


class NotAComparableHistoryError(Exception):
    r"""Generates an error if it is not possible to compare the values in the
    history."""


class HistoryAllCloseOperator(BaseAllCloseOperator[BaseHistory]):
    r"""Implements an allclose operator for ``BaseHistory`` objects."""

    def allclose(
        self,
        tester: BaseAllCloseTester,
        object1: BaseHistory,
        object2: Any,
        rtol: float = 1e-5,
        atol: float = 1e-8,
        equal_nan: bool = False,
        show_difference: bool = False,
    ) -> bool:
        if not isinstance(object2, BaseHistory):
            if show_difference:
                logger.info(f"object2 is not a `BaseHistory` object: {type(object2)}")
            return False
        object_equal = object1.equal(object2)
        if show_difference and not object_equal:
            logger.info(
                f"`BaseHistory` objects are different\nobject1=\n{object1}\nobject2=\n{object2}"
            )
        return object_equal


class HistoryEqualityOperator(BaseEqualityOperator[BaseHistory]):
    r"""Implements an equality operator for ``BaseHistory`` objects."""

    def equal(
        self,
        tester: BaseEqualityTester,
        object1: BaseHistory,
        object2: Any,
        show_difference: bool = False,
    ) -> bool:
        if not isinstance(object2, BaseHistory):
            if show_difference:
                logger.info(f"object2 is not a `BaseHistory` object: {type(object2)}")
            return False
        object_equal = object1.equal(object2)
        if show_difference and not object_equal:
            logger.info(
                f"`BaseHistory` objects are different\nobject1=\n{object1}\nobject2=\n{object2}"
            )
        return object_equal


if not AllCloseTester.has_allclose_operator(BaseHistory):
    AllCloseTester.add_allclose_operator(BaseHistory, HistoryAllCloseOperator())  # pragma: no cover
if not EqualityTester.has_equality_operator(BaseHistory):
    EqualityTester.add_equality_operator(BaseHistory, HistoryEqualityOperator())  # pragma: no cover



r"""This module implements some utility functions to use ``torch.Tensor``s."""

__all__ = ["LazyFlattedTensor"]

from typing import Any, Optional

import torch
from torch import Tensor

from gravitorch.distributed.ddp import all_gather_tensor_varshape
from gravitorch.utils.format import str_add_indent


class LazyFlattedTensor:
    r"""Implements a class to lazily concatenate flatted tensors.

    This class is at a very early stage and is very likely to change
    a lot in the future.

    To be more efficient, the tensors are concatenated only when the
    method ``consolidate`` is called. The tensors are stored in an
    internal buffer, then they are concatenated and stored in a
    separate variable. Storing the result in a separate variable
    leads to a more efficient design because the tensor is generated
    only one time. Adding another tensor leads to a new tensor when
    the method ``consolidate`` is called.

    Note: this class is independent of LazyTensor (see
    https://pytorch.org/blog/understanding-lazytensor-system-performance-with-pytorch-xla-on-cloud-tpu/
    for more information).

    Args:
        values (``torch.Tensor`` or ``None``, optional): Specifies the
            initial values. The tensor is flattened if necessary.
            ``None`` means no initial values. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.tensor import LazyFlattedTensor
        >>> lazy_tensor = LazyFlattedTensor()
        >>> lazy_tensor.update(torch.arange(6))
        >>> lazy_tensor.update(torch.tensor([-3, 1, 7]))
        >>> lazy_tensor.values()
        tensor([ 0.,  1.,  2.,  3.,  4.,  5., -3.,  1.,  7.])
        >>> lazy_tensor.update(torch.arange(3))
        >>> lazy_tensor.values()
        tensor([ 0.,  1.,  2.,  3.,  4.,  5., -3.,  1.,  7.,  0.,  1.,  2.])
        # By default, the tensor type is torch.float32. To use another type like long,
        # you need to specify the target type when creating the LazyFlattedTensor object.
        >>> lazy_tensor = LazyFlattedTensor(torch.tensor([], dtype=torch.long))
        >>> lazy_tensor.update(torch.arange(6))
    """

    def __init__(self, values: Optional[Tensor] = None):
        if values is None:
            values = torch.tensor([])
        self._values = values.flatten()
        self._buffer = []

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  values={str_add_indent(self._values)}\n"
            f"  buffer={str_add_indent(self._buffer)}\n)"
        )

    def all_reduce(self) -> "LazyFlattedTensor":
        r"""Reduces the values across all machines in such a way that all get
        the all the values.

        Returns:
            ``LazyFlattedTensor``: The reduced flatted tensor.

        Example usage:

        .. code-block:: python

            >>> import torch
            >>> from gravitorch.utils.tensor import LazyFlattedTensor
            >>> lazy_tensor = LazyFlattedTensor()
            >>> lazy_tensor.update(torch.arange(6))
            >>> lazy_tensor_reduced = lazy_tensor.all_reduce()
        """
        return LazyFlattedTensor(torch.cat(all_gather_tensor_varshape(self.values()), dim=0))

    def clear(self) -> None:
        r"""Clears the values and the internal buffer."""
        self._values = torch.tensor([])
        self._buffer.clear()

    def clone(self) -> "LazyFlattedTensor":
        r"""Creates a copy of the current lazy flatted tensor.

        Returns:
            ``LazyFlattedTensor``: A copy of the current lazy flatted
                tensor.
        """
        return LazyFlattedTensor(self.values().clone())

    def consolidate(self) -> None:
        r"""Consolidates the current values and internal buffer in a single
        flatted tensor.

        This method does nothing if the lazy tensor is already
        consolidated.
        """
        if self._buffer:
            values = self._values
            if values.numel() == 0:
                # Use the first tensor in the buffer to find the initial data type.
                values = values.to(dtype=self._buffer[0].dtype)
            self._values = torch.cat(
                [values] + [tensor.flatten() for tensor in self._buffer],
                dim=0,
            )
            self._buffer.clear()

    def equal(self, other: Any) -> bool:
        r"""Indicates if two lazy flatted tensors are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the two lazy flatted tensors are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, LazyFlattedTensor):
            return False
        if self.values().dtype != other.values().dtype:
            return False
        return self.values().equal(other.values())

    def numel(self) -> int:
        r"""Gets the total number of elements.

        Returns:
            int: The total number of elements.
        """
        return self._values.numel() + sum([tensor.numel() for tensor in self._buffer])

    def update(self, tensor: Tensor) -> None:
        r"""Updates the internal buffer by adding a new tensor.

        Args:
            tensor (``torch.Tensor``): Specifies the new tensor to add
                to the internal buffer. The tensor is flatted if
                necessary.
        """
        self._buffer.append(tensor)

    def values(self) -> Tensor:
        r"""Gets a flatted tensor with all the values.

        Returns:
            ``torch.Tensor``: The flatted tensor with all the values.
        """
        self.consolidate()
        return self._values



__all__ = ["EpochCudaMemoryMonitor", "IterationCudaMemoryMonitor"]

import torch

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.cudamem import log_max_cuda_memory_allocated
from gravitorch.utils.events import (
    ConditionalEventHandler,
    EpochPeriodicCondition,
    IterationPeriodicCondition,
)
from gravitorch.utils.exp_trackers import EpochStep, IterationStep


class EpochCudaMemoryMonitor(BaseHandler):
    r"""Implements a handler to monitor the CUDA memory usage every ``freq``
    epochs.

    Args:
        event (str, optional): Specifies the epoch-based event when
            the CUDA memory usage should be capture.
            Default: ``'epoch_completed'``
        freq (int, optional): Specifies the epoch frequency used to
            monitor the CUDA memory usage. Default: ``1``
    """

    def __init__(self, event: str = EngineEvents.EPOCH_COMPLETED, freq: int = 1):
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq}, event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=ConditionalEventHandler(
                self.monitor,
                condition=EpochPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def monitor(self, engine: BaseEngine) -> None:
        r"""Monitors the CUDA memory usage.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        log_max_cuda_memory_allocated()
        if torch.cuda.is_available():
            allocated_memory = torch.cuda.max_memory_allocated()
            total_memory = torch.cuda.mem_get_info()[1]
            engine.log_metrics(
                {
                    "epoch/max_cuda_memory_allocated": allocated_memory,
                    "epoch/max_cuda_memory_allocated_pct": float(allocated_memory / total_memory),
                },
                step=EpochStep(engine.epoch),
            )


class IterationCudaMemoryMonitor(BaseHandler):
    r"""Implements a handler to monitor the CUDA memory usage every ``freq``
    iterations.

    Args:
        event (str, optional): Specifies the iteration-based event
            when the CUDA memory usage should be capture.
            Default: ``'epoch_completed'``
        freq (int, optional): Specifies the iteration frequency used
            to monitor the CUDA memory usage. Default: ``1``
    """

    def __init__(self, event: str = EngineEvents.TRAIN_ITERATION_COMPLETED, freq: int = 1):
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq}, event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=ConditionalEventHandler(
                self.monitor,
                condition=IterationPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def monitor(self, engine: BaseEngine) -> None:
        r"""Monitors the CUDA memory usage.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        log_max_cuda_memory_allocated()
        if torch.cuda.is_available():
            allocated_memory = torch.cuda.max_memory_allocated()
            total_memory = torch.cuda.mem_get_info()[1]
            engine.log_metrics(
                {
                    "iteration/max_cuda_memory_allocated": allocated_memory,
                    "iteration/max_cuda_memory_allocated_pct": float(
                        allocated_memory / total_memory
                    ),
                },
                step=IterationStep(engine.iteration),
            )



r"""This module defines a metric to compute the logarithm of the hyperbolic
cosine of the prediction error."""

__all__ = ["LogCoshError"]

import logging
from typing import Union

from torch import Tensor

from gravitorch.models.metrics.base_epoch import BaseStateEpochMetric
from gravitorch.models.metrics.state import BaseState, ErrorState
from gravitorch.nn.functional import log_cosh_loss

logger = logging.getLogger(__name__)


class LogCoshError(BaseStateEpochMetric):
    r"""Implements a metric to compute the logarithm of the hyperbolic cosine of
    the prediction error.

    Args:
        mode (str): Specifies the mode (e.g ``'train'`` or ``'eval'``).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'log_cosh_err'``
        scale (float, optional): Specifies the scale factor.
            Default: ``1.0``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import LogCoshError
        >>> metric = LogCoshError('eval')
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/log_cosh_err_mean': 0.0,
         'eval/log_cosh_err_min': 0.0,
         'eval/log_cosh_err_max': 0.0,
         'eval/log_cosh_err_sum': 0.0,
         'eval/log_cosh_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/log_cosh_err_mean': 0.07229681313037872,
         'eval/log_cosh_err_min': 0.0,
         'eval/log_cosh_err_max': 0.43378087878227234,
         'eval/log_cosh_err_sum': 0.8675617575645447,
         'eval/log_cosh_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/log_cosh_err_mean': 0.21689043939113617,
         'eval/log_cosh_err_min': 0.0,
         'eval/log_cosh_err_max': 0.43378087878227234,
         'eval/log_cosh_err_sum': 0.8675617575645447,
         'eval/log_cosh_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "log_cosh_err",
        scale: float = 1.0,
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or ErrorState())
        if scale <= 0.0:
            raise ValueError(f"Incorrect scale {scale}. The scale has to be >0")
        self._scale = float(scale)

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the metric given a mini-batch of examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.
        """
        self._state.update(log_cosh_loss(prediction, target, reduction="none", scale=self._scale))



r"""This module implements a generic history."""

__all__ = ["GenericHistory"]

from collections import deque
from collections.abc import Iterable
from typing import Any, Optional, TypeVar

from coola import objects_are_equal

from gravitorch.utils.history.base import BaseHistory, EmptyHistoryError

T = TypeVar("T")


class GenericHistory(BaseHistory[T]):
    r"""Implements a generic history to store the recent values added in the
    history.

    Internally, this class uses a ``deque`` to keep the most recent
    values added in the history. Note that this class does not allow
    to get the best value because it is not possible to define a
    generic rule to know the best object. Please see
    ``ScalarHistory`` that can compute the best value for
    scalars.

    Args:
        name (str): Specifies the name of the history.
        elements (iterable, optional): Specifies the initial elements.
            Each element is a tuple with the step and its associated
            value. Default: ``tuple()``
        max_size (int, optional): Specifies the maximum size
            of the history. Default: ``10``
    """

    def __init__(
        self,
        name: str,
        elements: Iterable[tuple[Optional[int], T]] = tuple(),
        max_size: int = 10,
    ):
        super().__init__(name)
        if max_size <= 0:
            raise ValueError(f"History size must be greater than 0! (received: {max_size})")
        self._history = deque(elements, maxlen=max_size)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(name={self.name}, "
            f"max_size={self.max_size}, history={self.get_recent_history()})"
        )

    @property
    def max_size(self) -> int:
        r"""int: The maximum size of the history"""
        return self._history.maxlen

    def add_value(self, value: T, step: Optional[int] = None) -> None:
        self._history.append((step, value))

    def equal(self, other: Any) -> bool:
        if not isinstance(other, GenericHistory):
            return False
        return objects_are_equal(self.to_dict(), other.to_dict())

    def get_last_value(self) -> Any:
        if self.is_empty():
            raise EmptyHistoryError(
                f"It is not possible to get the last value because the history {self._name} "
                "is empty"
            )
        return self._history[-1][1]

    def get_recent_history(self) -> tuple[tuple[Optional[int], T], ...]:
        return tuple(self._history)

    def is_comparable(self) -> bool:
        return False

    def is_empty(self) -> bool:
        return not self._history

    def config_dict(self) -> dict[str, Any]:
        config = super().config_dict()
        config["max_size"] = self.max_size
        return config

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        self._history = deque(state_dict["history"], maxlen=self.max_size)

    def state_dict(self) -> dict[str, Any]:
        return {"history": self.get_recent_history()}



r"""This module implements an evaluation loop using automatic mixed precision
(AMP)."""

__all__ = ["AMPEvaluationLoop"]

import logging
from typing import Any, Union

import torch
from torch.cuda.amp import autocast
from torch.nn import Module

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.utils.device_placement import BaseDevicePlacement
from gravitorch.utils.evaluation_loops.conditions import BaseEvalCondition
from gravitorch.utils.evaluation_loops.vanilla import VanillaEvaluationLoop
from gravitorch.utils.loop_observers import BaseLoopObserver
from gravitorch.utils.profilers import BaseProfiler

logger = logging.getLogger(__name__)


class AMPEvaluationLoop(VanillaEvaluationLoop):
    r"""Implements a training loop to train a model on a dataset by using
    training loop using automatic mixed precision (AMP).

    Args:
        grad_enabled (bool, optional): Specifies if the gradient is
            computed or not in the evaluation loop. By default, the
            gradient is not computed to reduce the memory footprint.
            Default: ``False``
        amp_enabled (bool, optional): If ``True``, automatic mixed
            precision (AMP) is enabled, otherwise it is disabled.
            Default: ``True``
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"eval"``
        condition (``BaseEvalCondition`` or dict or None): Specifies
            the condition to evaluate the loop or its configuration.
            If ``None``, the ``EveryEpochEvalCondition(every=1)`` is
            used.  Default ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional):
            Specifies the profiler or its configuration. If ``None``,
            the ``NoOpProfiler`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        tag: str = "eval",
        grad_enabled: bool = False,
        amp_enabled: bool = True,
        batch_device_placement: Union[BaseDevicePlacement, dict, None] = None,
        condition: Union[BaseEvalCondition, dict, None] = None,
        observer: Union[BaseLoopObserver, dict, None] = None,
        profiler: Union[BaseProfiler, dict, None] = None,
    ):
        super().__init__(
            tag=tag,
            grad_enabled=grad_enabled,
            batch_device_placement=batch_device_placement,
            condition=condition,
            observer=observer,
            profiler=profiler,
        )
        self._amp_enabled = bool(amp_enabled)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  tag={self._tag},\n"
            f"  batch_device_placement={self._batch_device_placement},\n"
            f"  grad_enabled={self._grad_enabled},\n"
            f"  amp_enabled={self._amp_enabled},\n"
            f"  condition={self._condition},\n"
            f"  observer={self._observer},\n"
            f"  profiler={self._profiler},\n"
            ")"
        )

    def _eval_one_batch(self, engine: BaseEngine, model: Module, batch: Any) -> dict:
        engine.fire_event(EngineEvents.EVAL_ITERATION_STARTED)
        with torch.set_grad_enabled(self._grad_enabled):
            with autocast(enabled=self._amp_enabled):
                output = model(self._batch_device_placement.send(batch))
        engine.fire_event(EngineEvents.EVAL_ITERATION_COMPLETED)
        return output



r"""This module defines a criterion wrapper to make PyTorch criteria compatible
with ``gravitorch.models.VanillaModel`` and padded sequences."""

__all__ = ["PaddedSequenceLoss"]

from typing import Union

import torch
from torch import Tensor
from torch.nn import Module

from gravitorch import constants as ct
from gravitorch.nn import setup_nn_module


class PaddedSequenceLoss(Module):
    r"""Implements a wrapper to adapt PyTorch loss function (a.k.a. criterion)
    to deal with zero padded sequential inputs.

    This class works for most of the PyTorch criterion that has two
    inputs: prediction and target. It assumes that the valid time
    steps of the sequence are indicated with a mask. This loss
    function should have at least two inputs:

        - the prediction which is a ``torch.Tensor`` of shape
            ``(sequence_length, batch_size, *)`` or
            ``(batch_size, sequence_length, *)`` where ``*`` means
            any number of additional dimensions. This tensor is
            converted to a tensor of shape
            ``(sequence_length * batch_size, *)`` and then feeds to
            the PyTorch loss function.
        - the target which is a ``torch.Tensor`` of shape
            ``(sequence_length, batch_size, *)`` or
            ``(batch_size, sequence_length, *)`` where ``*`` means
            any number of additional dimensions.
            This tensor is converted to a tensor of shape
            ``(sequence_length * batch_size, *)`` and then feeds to
            the PyTorch loss function.

    The input mask is optional. If no mask is provided, all the steps
    are considered as valid. The mask is a ``torch.Tensor`` of shape
    ``(sequence_length, batch_size)`` or
    ``(batch_size, sequence_length)``. The type of the tensor can be
    ``torch.int`` or``torch.long`` or``torch.float`` or ``torch.bool``
    with the following values:

        - valid value: ``True`` or ``1`` if ``valid_value=True``,
            otherwise ``False`` or ``0``.
        - invalid value: ``False`` or ``0`` if ``valid_value=True``,
            otherwise ``True`` or ``1``.

    Note that this class may not be compatible which any PyTorch
    criterion. However, you should be able to adapt this
    implementation for your use-case.

    Args:
        criterion (``torch.nn.Module`` or dict): Specifies the loss
            function or its configuration.
        prediction_key (str): Specifies the prediction key.
            Default: ``"prediction"``.
        target_key (str): Specifies the target key.
            Default: ``"target"``.
        mask_key (str): Specifies the mask key. Default: ``"mask"``.
        valid_value (bool, optional): Indicates the valid values in
            the mask. If ``True``, the valid values are indicated by
            a ``True`` in the mask. If ``False``, the valid values are
            indicated by a ``False`` in the mask.
            Default: ``True``
        mask_in_batch (bool, optional): Indicates if the mask is in
            ``batch`` or ``net_out``. If ``True``, the mask is taken
            from the input ``batch``, otherwise it is taken from the
            input ``net_out``. Default: ``True``

    Example usage:

    .. code-block:: python

        >>> from torch import nn
        >>> from gravitorch.models.criterions import PaddedSequenceLoss
        # Init with a nn.Module
        >>> criterion = PaddedSequenceLoss(criterion=nn.MSELoss())
        # Init with a config
        >>> criterion = PaddedSequenceLoss(criterion={'_target_': 'torch.nn.MSELoss'})
        # Customize keys.
        >>> criterion = PaddedSequenceLoss(
        ...     criterion=nn.MSELoss(),
        ...     prediction_key='my_prediction',
        ...     target_key='my_target',
        ...     mask_key='my_mask',
        ... )
        >>> net_out = {'my_prediction': ...}
        >>> batch = {'my_target': ..., 'my_mask': ...}
        >>> loss = criterion(net_out, batch)
        {'loss': torch.tensor(...)}
    """

    def __init__(
        self,
        criterion: Union[Module, dict],
        prediction_key: str = ct.PREDICTION,
        target_key: str = ct.TARGET,
        mask_key: str = ct.MASK,
        valid_value: bool = True,
        mask_in_batch: bool = True,
    ):
        super().__init__()
        self.criterion = setup_nn_module(criterion)
        self._prediction_key = prediction_key
        self._target_key = target_key
        self._mask_key = mask_key
        self._valid_value = bool(valid_value)
        self._mask_in_batch = bool(mask_in_batch)

    def forward(self, net_out: dict, batch: dict) -> dict[str, Tensor]:
        r"""Computes the loss value given the network output and the batch.

        Args:
            net_out (dict ): Specifies the network output which
                contains the prediction.
            batch (dict): Specifies the batch which contains the
                target and the mask.

        Returns:
            dict: A dict with the loss value.
        """
        # See the batch of sequences as a batch of examples
        prediction = self._get_prediction_from_net_out(net_out).flatten(0, 1)
        target = self._get_target_from_batch(batch).flatten(0, 1)

        # Get the mask and remove the examples that are masked
        mask = (
            batch.get(self._mask_key, None)
            if self._mask_in_batch
            else net_out.get(self._mask_key, None)
        )
        if mask is not None:
            mask = mask.flatten().bool()
            if not self._valid_value:
                mask = torch.logical_not(mask)
            prediction = prediction[mask]
            target = target[mask]

        return {ct.LOSS: self.criterion(prediction, target)}

    def _get_prediction_from_net_out(self, net_out: dict) -> Tensor:
        r"""Gets the prediction from the network output.

        Args:
            net_out (dict): Specifies the network output which
                contains the prediction.

        Returns:
            ``torch.Tensor``: the prediction
        """
        return net_out[self._prediction_key]

    def _get_target_from_batch(self, batch: dict) -> Tensor:
        r"""Gets the target from the batch. The target is the tensor with the
        key 'target'.

        Args:
            batch (dict): Specifies the batch which contains the
                target.

        Returns:
            ``torch.Tensor``: the target
        """
        return batch[self._target_key]



__all__ = ["ResidualBlock"]

from typing import Any, Union

from torch import Tensor
from torch.nn import Identity, Module

from gravitorch.nn.utils import setup_nn_module


class ResidualBlock(Module):
    r"""Implementation of a residual block.

    Args:
        residual (``torch.nn.Module`` or dict): Specifies the residual
            mapping module or its configuration (dictionary).
        skip (``torch.nn.Module`` or dict or ``None``): Specifies the
            skip mapping module or its configuration (dictionary). If
            ``None``, the ``Identity`` module is used.
            Default: ``None``
    """

    def __init__(
        self,
        residual: Union[Module, dict[str, Any]],
        skip: Union[Module, dict[str, Any], None] = None,
    ):
        super().__init__()
        self.residual = setup_nn_module(residual)
        self.skip = setup_nn_module(skip or Identity())

    def forward(self, tensor: Tensor) -> Tensor:
        return self.skip(tensor) + self.residual(tensor)



__all__ = ["DDPPartitioner", "SyncParallelPartitioner"]

from collections.abc import Sequence
from typing import Optional, TypeVar, Union

from gravitorch.data.partitioners.base import BasePartitioner, setup_partitioner
from gravitorch.distributed.ddp import broadcast_object_list
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_add_indent
from gravitorch.utils.partitioning import ddp_partitions

T = TypeVar("T")


class DDPPartitioner(BasePartitioner[T]):
    r"""Implements a partitioner designed for a Distributed Data Parallel (DDP)
    setting.

    All the partitions have the same number of items. The number of
    partitions is the world size of the distributed system.

    Args:
        partition_size (int): Specifies the partition size i.e. the
            number of items in each partition.
    """

    def __init__(self, partition_size: int):
        self._partition_size = int(partition_size)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(partition_size={self.partition_size:,})"

    @property
    def partition_size(self) -> int:
        r"""int: The partition size."""
        return self._partition_size

    def partition(
        self, items: Sequence[T], engine: Optional[BaseEngine] = None
    ) -> list[Sequence[T]]:
        return ddp_partitions(items=items, partition_size=self._partition_size)


class SyncParallelPartitioner(BasePartitioner[T]):
    r"""Implements a partitioner that synchronize the partitions in a
    distributed setting.

    This partitioner ensures all the distributed processes have the
    same partitions. The partitions from the main process (rank 0)
    are broadcast to the other processes.

    Args:
        partitioner (``BasePartitioner`` or dict): Specifies a
            partitioner or its configuration.
    """

    def __init__(self, partitioner: Union[BasePartitioner, dict]):
        self._partitioner = setup_partitioner(partitioner)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  partitioner={str_add_indent(self._partitioner)},\n"
            ")"
        )

    @property
    def partitioner(self) -> BasePartitioner:
        r"""``BasePartitioner``: The partitioner."""
        return self._partitioner

    def partition(
        self, items: Sequence[T], engine: Optional[BaseEngine] = None
    ) -> list[Sequence[T]]:
        partitions = self._partitioner.partition(items, engine)
        # Synchronize the partitions between process so all the distributed processes
        # have the same partitions.
        broadcast_object_list(partitions)
        return partitions



r"""This module defines the base class for the model parameter
parameter_initializers."""

__all__ = ["BaseParameterInitializer", "BaseDefaultParameterInitializer"]

import logging
from abc import ABC, abstractmethod

from objectory import AbstractFactory

from gravitorch.engines.base import BaseEngine
from gravitorch.nn.utils.parameter_analysis import show_parameter_stats

logger = logging.getLogger(__name__)


class BaseParameterInitializer(ABC, metaclass=AbstractFactory):
    r"""Defines the base parameter initializer.

    Note that there are other ways to initialize the model parameters.
    For example, you can initialize the weights of your model directly
    in the model.
    """

    @abstractmethod
    def initialize(self, engine: BaseEngine) -> None:
        r"""Initializes the parameters of the model.

        Note that the input is the engine instead of the model because
        it allows to initialize the weights based on the dataset or
        other parameters in the engine. The parameters of the model
        should be updated in-place.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """


class BaseDefaultParameterInitializer(BaseParameterInitializer):
    r"""Defines a base parameter initializer that shows some stats about the
    parameters.

    Note that there are other ways to initialize the model parameters.
    For example, you can initialize the weights of your model directly
    in the model.

    Args:
        show_stats (bool, optional): If ``True``, the parameter
            statistics are shown at the end of the initialization.
            Default: ``True``
    """

    def __init__(self, show_stats: bool = True):
        self._show_stats = show_stats

    def initialize(self, engine: BaseEngine) -> None:
        r"""Initializes the parameters of the model and shows some stats.

        Note that the input is the engine instead of the model because
        it allows to initialize the weights based on the dataset or
        other parameters in the engine. The parameters of the model
        should be updated in-place.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        self._initialize(engine)
        if self._show_stats:
            show_parameter_stats(engine.model)

    @abstractmethod
    def _initialize(self, engine: BaseEngine) -> None:
        r"""Initializes the parameters of the model.

        Note that the input is the engine instead of the model because
        it allows to initialize the weights based on the dataset or
        other parameters in the engine. The parameters of the model
        should be updated in-place.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """



__all__ = ["DatasetDataSource"]

import logging
from collections.abc import Iterable
from typing import Any, Optional, TypeVar, Union

from torch.utils.data import Dataset

from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.creators.dataloader.utils import setup_data_loader_creator
from gravitorch.data.datasets import setup_dataset
from gravitorch.datasources.base import BaseDataSource, LoaderNotFoundError
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.asset_manager import AssetManager
from gravitorch.utils.format import str_add_indent, to_torch_mapping_str

logger = logging.getLogger(__name__)

T = TypeVar("T")


class DatasetDataSource(BaseDataSource):
    r"""Implements a data source that uses regular PyTorch datasets and data
    loaders.

    To create a data loader, the user should indicate the dataset and
    the data loader creator. Note that the regular PyTorch data
    loader needs a ``Dataset`` object has input.

    Args:
        datasets (dict): Specifies the datasets to initialize. Each
            key indicates the dataset name. It is possible to give a
            ``Dataset`` object, or the configuration of a ``Dataset``
            object.
        data_loader_creators (dict): Specifies the data loader
            creators to initialize. Each key indicates a data loader
            creator name. For example if you want to create a data
            loader for ``'train'`` ID, the dictionary has to have a
            key ``'train'``. The value can be a
            ``BaseDataLoaderCreator`` object, or its configuration,
            or ``None``. ``None`` means a default data loader will be
            created. Each data loader creator takes a ``Dataset``
            object as input, so you need to specify a dataset with the
            same name.
    """

    def __init__(
        self,
        datasets: dict[str, Union[Dataset, dict]],
        data_loader_creators: dict[str, Union[BaseDataLoaderCreator, dict, None]],
    ):
        self._asset_manager = AssetManager()

        logger.info("Initializing the datasets...")
        self._datasets = {key: setup_dataset(dataset) for key, dataset in datasets.items()}
        logger.info(f"datasets:\n{to_torch_mapping_str(self._datasets)}")

        logger.info("Initializing the data loader creators...")
        self._data_loader_creators = {
            key: setup_data_loader_creator(creator) for key, creator in data_loader_creators.items()
        }
        logger.info(f"data loader creators:\n{to_torch_mapping_str(self._data_loader_creators)}")
        self._check()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            "  datasets:\n"
            f"    {str_add_indent(to_torch_mapping_str(self._datasets), num_spaces=4)}\n"
            "  data_loader_creators:\n"
            f"    {str_add_indent(to_torch_mapping_str(self._data_loader_creators), num_spaces=4)}"
            "\n)"
        )

    def attach(self, engine: BaseEngine) -> None:
        logger.info("Attach the data source to an engine")

    def get_asset(self, asset_id: str) -> Any:
        return self._asset_manager.get_asset(asset_id)

    def has_asset(self, asset_id: str) -> bool:
        return self._asset_manager.has_asset(asset_id)

    def get_data_loader(self, loader_id: str, engine: Optional[BaseEngine] = None) -> Iterable[T]:
        if not self.has_data_loader(loader_id):
            raise LoaderNotFoundError(f"{loader_id} does not exist")
        return self._data_loader_creators[loader_id].create(
            dataset=self._datasets[loader_id], engine=engine
        )

    def has_data_loader(self, loader_id: str) -> bool:
        return loader_id in self._data_loader_creators

    def _check(self) -> None:
        # Verify each data loader creator has a dataset
        for key in self._data_loader_creators:
            if key not in self._datasets:
                logger.warning(f"Missing '{key}' dataset for its associated data loader creator")
        # Verify each dataset has a data loader creator
        for key in self._datasets:
            if key not in self._data_loader_creators:
                logger.warning(f"Missing '{key}' data loader creator for its associated dataset")



r"""This module implements a PyTorch profiler."""

__all__ = ["PyTorchProfiler"]

import logging

import torch

from gravitorch.utils.profilers.base import BaseProfiler

logger = logging.getLogger(__name__)


class PyTorchProfiler(BaseProfiler):
    r"""Implements a PyTorch profiler.

    This profiler relies on the package ``torch.profiler`` to profile
    the code.

    Args:
        profiler (``torch.profiler.profile``): Specifies the profiler
            to use.
    """

    def __init__(self, profiler: torch.profiler.profile):
        self._profiler = profiler

    def __enter__(self):
        logger.info("Starting PyTorch profiler...")
        self._profiler.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.info("Ending PyTorch profiler")
        self._profiler.__exit__(exc_type, exc_val, exc_tb)

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def step(self) -> None:
        r"""Signals the profiler that the next profiling step has started."""
        self._profiler.step()

    @classmethod
    def scheduled_profiler_with_tensorboard_trace(
        cls,
        trace_path: str,
        wait: int,
        warmup: int,
        active: int,
        repeat: int = 0,
        skip_first: int = 0,
        record_shapes: bool = False,
        profile_memory: bool = False,
        with_stack: bool = False,
        with_flops: bool = False,
    ) -> "PyTorchProfiler":
        r"""Implements a scheduled profiler with a TensorBoard trace.

        The profiler will skip the first ``skip_first`` steps, then
        wait for ``wait`` steps, then do the warmup for the next
        ``warmup`` steps, then do the active recording for the next
        ``active`` steps and then repeat the cycle starting with
        ``wait`` steps. The optional number of cycles is specified
        with the ``repeat`` parameter, the zero value means that the
        cycles will continue until the profiling is finished.

        Args:
            trace_path (str): Specifies the path where to write the
                profiling trace. This path can be directly delivered
                to TensorBoard as logdir.
            wait (int): Specifies the number of waiting steps in a
                cycle.
            warmup (int): Specifies the number of warmup steps in a
                cycle.
            active (int): Specifies the number of active steps in a
                cycle.
            repeat (int, optional): Specifies the number of cycles.
                If ``0``, the cycles will continue until the profiling
                is finished. Default: ``0``
            skip_first (int, optional): Specifies the number of steps
                that are skipped at the beginning. Default: ``0``
            record_shapes (bool, optional): If ``True``, the profiler
                saves information about operator’s input shapes.
                Default: ``False``
            profile_memory (bool, optional): If ``True``, the profiler
                tracks tensor memory allocation/deallocation.
                Default: ``False``
            with_stack (bool, optional): If ``True``, the profiler
                record source information (file and line number) for
                the ops. Default: ``False``
            with_flops (bool, optional): If ``True``, the profiler
                uses formula to estimate the FLOPS of specific
                operators (matrix multiplication and 2D convolution).
                Default: ``False``

        Returns:
            ``PyTorchProfiler``: A scheduled profiler with a TensorBoard trace.
        """
        return cls(
            torch.profiler.profile(
                schedule=torch.profiler.schedule(
                    wait=wait,
                    warmup=warmup,
                    active=active,
                    repeat=repeat,
                    skip_first=skip_first,
                ),
                on_trace_ready=torch.profiler.tensorboard_trace_handler(trace_path),
                record_shapes=record_shapes,
                profile_memory=profile_memory,
                with_stack=with_stack,
                with_flops=with_flops,
            )
        )



r"""This module defines the base class for the training loops."""

__all__ = ["BaseTrainingLoop", "BaseBasicTrainingLoop"]

import logging
import sys
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from typing import Any, Optional, Union

from objectory import AbstractFactory
from torch.nn import Module
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.optimizers.utils import (
    log_optimizer_parameters_per_group,
    show_optimizer_parameters_per_group,
)
from gravitorch.utils.cudamem import log_max_cuda_memory_allocated
from gravitorch.utils.exp_trackers import EpochStep
from gravitorch.utils.history import MinScalarHistory
from gravitorch.utils.loop_observers import BaseLoopObserver, setup_loop_observer
from gravitorch.utils.metric_tracker import ScalarMetricTracker
from gravitorch.utils.profilers import BaseProfiler, setup_profiler
from gravitorch.utils.seed import manual_seed
from gravitorch.utils.timing import BatchLoadingTimer

logger = logging.getLogger(__name__)


class BaseTrainingLoop(ABC, metaclass=AbstractFactory):
    r"""Defines the training loop base class.

    To implement your own training loop, you will need to define the
    ``train`` method. If the training loop has state values, you also
    need to implement the following methods:

        - ``load_state_dict``
        - ``state_dict``
    """

    @abstractmethod
    def train(self, engine: BaseEngine) -> None:
        r"""Trains the model on the training dataset.

        The training metrics/artifacts should be logged through the
        engine.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Sets up the training loop from a dictionary containing the state
        values.

        Args:
            state_dict (dict): Specifies a dictionary
                containing state keys with values.
        """

    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {}


class BaseBasicTrainingLoop(BaseTrainingLoop):
    r"""Implements a base class to implement basic training loops.

    Child classes have to implement the following methods:

        - ``_prepare_model_optimizer_data_loader``
        - ``_setup_clip_grad``
        - ``_train_one_batch``

    This class was implemented to reduce the duplicated code between
    ``VanillaTrainingLoop`` and ``AccelerateTrainingLoop``.

    Args:
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"train"``
        clip_grad (dict or None, optional): Specifies the
            configuration to clip the gradient. If ``None``, no
            gradient clipping is used during the training.
            Default: ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional):
            Specifies the profiler or its configuration.
            If ``None``, the ``NoOpProfiler`` is instantiated.
            Default: ``None``
    """

    def __init__(
        self,
        tag: str = "train",
        clip_grad: Optional[dict] = None,
        observer: Union[BaseLoopObserver, dict, None] = None,
        profiler: Union[BaseProfiler, dict, None] = None,
    ):
        self._tag = str(tag)
        self._clip_grad_fn, self._clip_grad_args = self._setup_clip_grad(clip_grad or {})
        self._observer = setup_loop_observer(observer)
        self._profiler = setup_profiler(profiler)
        logger.info(f"profiler:\n{self._profiler}")

    def train(self, engine: BaseEngine) -> None:
        r"""Trains the model on the training dataset.

        The training metrics/artifacts should be logged through the
        engine.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        dist.barrier()
        self._prepare_training(engine)
        engine.fire_event(EngineEvents.TRAIN_EPOCH_STARTED)

        model, optimizer, data_loader = self._prepare_model_optimizer_data_loader(engine)

        # Train the model on each mini-match in the dataset.
        metrics = ScalarMetricTracker()
        data_loader = BatchLoadingTimer(data_loader, epoch=engine.epoch, prefix=f"{self._tag}/")
        self._observer.start(engine)
        dist.barrier()

        with self._profiler as profiler:
            for batch in data_loader:
                engine.increment_iteration()
                # Run forward/backward on the given batch.
                output = self._train_one_batch(engine, model, optimizer, batch)
                metrics.update(output)
                self._observer.update(engine=engine, model_input=batch, model_output=output)
                profiler.step()

        # To be sure the progress bar is displayed before the following lines
        sys.stdout.flush()
        dist.barrier()
        self._observer.end(engine)

        # Log some training metrics to the engine.
        data_loader.log_stats(engine=engine)
        metrics.log_average_value(engine=engine, prefix=f"{self._tag}/")
        log_max_cuda_memory_allocated()
        dist.barrier()

        engine.fire_event(EngineEvents.TRAIN_EPOCH_COMPLETED)
        dist.barrier()

    def _prepare_training(self, engine: BaseEngine) -> None:
        r"""Prepares the training.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        logger.info(f"Preparing training for epoch {engine.epoch}...")
        # Fix the random seed for reproducibility purpose.
        manual_seed(engine.random_seed + engine.epoch + engine.max_epochs * dist.get_rank())
        engine.model.train()

        if not engine.has_history(f"{self._tag}/{ct.LOSS}"):
            engine.add_history(MinScalarHistory(f"{self._tag}/{ct.LOSS}"))

        show_optimizer_parameters_per_group(engine.optimizer)
        log_optimizer_parameters_per_group(
            optimizer=engine.optimizer,
            engine=engine,
            step=EpochStep(engine.epoch),
            prefix=f"{self._tag}/",
        )

    @abstractmethod
    def _prepare_model_optimizer_data_loader(
        self, engine: BaseEngine
    ) -> tuple[Module, Optimizer, Iterable]:
        r"""Prepares the model, optimizer and data loader.

        Args:
            engine (``BaseEngine``): Specifies the engine.

        Returns:
            ``torch.nn.Module``, ``torch.optim.Optimizer``,
                ``Iterable``: A tuple with the model, the optimizer
                and the data loader.
        """

    @abstractmethod
    def _setup_clip_grad(self, clip_grad: dict) -> tuple[Optional[Callable], tuple]:
        r"""Initializes the clipping gradient strategy during training.

        Args:
            clip_grad (dict): Specifies the clipping gradient option.

        Returns:
            tuple: clip gradient function, clip gradient arguments.

        Raises:
            ValueError: if it is an invalid clipping gradient option.
        """

    @abstractmethod
    def _train_one_batch(
        self, engine: BaseEngine, model: Module, optimizer: Optimizer, batch: Any
    ) -> dict:
        """Trains the model on the given batch.

        Args:
            engine (``BaseEngine``): Specifies the engine.
            model (``torch.nn.Module``): Specifies the model to train.
            optimizer (``torch.optim.optimizer``): Specifies the
                optimizer used to train the model.
            batch: Specifies the batch of data.

        Returns:
            dict: Some results (including the loss value) about the
                batch.
        """



__all__ = ["Logging", "LoggingState"]

import logging
from dataclasses import dataclass
from types import TracebackType
from typing import Optional, Union

from gravitorch.distributed import comm as dist
from gravitorch.rsrc.base import BaseResource

logger = logging.getLogger(__name__)


@dataclass
class LoggingState:
    disabled_level: int

    def restore(self) -> None:
        r"""Restores the logging configuration by using the values in the
        state."""
        logging.disable(self.disabled_level)

    @classmethod
    def create(cls) -> "LoggingState":
        r"""Creates a state to capture the current logging configuration.

        Returns:
            ``LoggingState``: The current logging state.
        """
        return cls(disabled_level=logging.root.manager.disable)


class Logging(BaseResource):
    r"""Implements a context manager to disable the logging.

    Args:
        only_main_process (bool, optional): If ``True``, only the
            outputs of the main process are logged. The logging of
            other processes is limited to the error level or above.
            If ``False``, the outputs of all the processes are logged.
            Default: ``True``
        disabled_level (int or str, optional): All logging calls
            of severity ``disabled_level`` and below will be
            disabled. Default: ``39``
        log_info (bool, optional): If ``True``, the state is shown
            after the context manager is created. Default: ``False``
    """

    def __init__(
        self, only_main_process: bool = False, disabled_level: Union[int, str] = logging.ERROR - 1
    ):
        self._only_main_process = bool(only_main_process)
        if isinstance(disabled_level, str):
            disabled_level = logging.getLevelName(disabled_level)
        self._disabled_level = int(disabled_level)

        self._state: list[LoggingState] = []

    def __enter__(self) -> "Logging":
        logger.info("Setting logging configuration...")
        self._state.append(LoggingState.create())
        if self._only_main_process and not dist.is_main_process():
            logging.disable(self._disabled_level)
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        logger.info("Restoring previous logging configuration...")
        self._state.pop().restore()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(only_main_process={self._only_main_process}, "
            f"disabled_level={self._disabled_level})"
        )



r"""This module implements a data summary for integer values."""

__all__ = ["IntegerDataSummary"]

from gravitorch.utils.data_summary.discrete import BaseDiscreteDataSummary


class IntegerDataSummary(BaseDiscreteDataSummary[int]):
    r"""Implements a data summary for integer values.

    This data summary assumes that the data are discrete numerical
    values. This data summary computes the following statistics:

        - ``count``: the number of values
        - ``num_unique_values``: the number of unique values
        - ``count_{}``: the number of values per unique value
    """

    def add(self, data: int) -> None:
        r"""Adds new data to the summary.

        Args:
            data (int): Specifies the data to add to the summary.
        """
        self._counter[int(data)] += 1



__all__ = ["concise_summary"]

from collections.abc import Mapping, Sequence
from itertools import islice
from typing import Any, Union

import numpy as np
import torch
from torch import Tensor

from gravitorch.utils.format import (
    str_add_indent,
    to_torch_mapping_str,
    to_torch_sequence_str,
)


def concise_summary(value: Any, num_spaces: int = 2, max_length: int = 5) -> str:
    r"""Creates a concise summary of a variable.

    This function supports the following types:

        - ``torch.Tensor``
        - ``numpy.ndarray``
        - ``list``
        - ``tuple``
        - ``dict``

    Args:
        value: Specifies the variable to summarize.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.
        max_length (int, optional): Specifies the number of values to
            show in the summary. This option is useful when the
            sequences/mappings are long because we may do not want to
            show all the values in the summary. Note this input is
            also used to create the summary of all the values in the
            sequence/mapping. Default: ``5``

    Returns:
        str: The concise summary of the variable.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.summary import concise_summary
        >>> print(concise_summary([1, 2, 3, 'abc', 5, 6, 7, 8, 9, 10]))
        <class 'list'> | length=10
          (0) <class 'int'> | value=1
          (1) <class 'int'> | value=2
          (2) <class 'int'> | value=3
          (3) <class 'str'> | value=abc
          (4) <class 'int'> | value=5
          ...
    """
    if torch.is_tensor(value) or isinstance(value, np.ndarray):
        return _concise_summary_tensor(value)
    if isinstance(value, (list, tuple)):
        return _concise_summary_sequence(
            sequence=value, num_spaces=num_spaces, max_length=max_length
        )
    if isinstance(value, dict):
        return _concise_summary_mapping(mapping=value, num_spaces=num_spaces, max_length=max_length)
    return f"{type(value)} | value={value}"


def _concise_summary_tensor(tensor: Union[Tensor, np.ndarray]) -> str:
    r"""Creates a concise summary of a ``torch.Tensor`` or ``numpy.ndarray``.

    Args:
       tensor: Specifies the tensor/array to summarize.

    Returns:
       str: The concise summary of the tensor/array.
    """
    summary = [f"{type(tensor)}", f"shape={tensor.shape}", f"dtype={tensor.dtype}"]
    if isinstance(tensor, np.ndarray):
        tensor = torch.from_numpy(tensor)
    if tensor.dtype in (torch.half, torch.float, torch.double):
        summary.extend(
            [
                f"mean={tensor.mean():,.6f}",
                f"std={tensor.std():,.6f}",
                f"min={tensor.min():,.6f}",
                f"max={tensor.max():,.6f}",
            ]
        )
    if tensor.dtype in (torch.int8, torch.short, torch.int, torch.long):
        summary.extend(
            [
                f"mean={tensor.float().mean():,.6f}",
                f"std={tensor.float().std():,.6f}",
                f"min={tensor.min()}",
                f"max={tensor.max()}",
            ]
        )
    return " | ".join(summary)


def _concise_summary_sequence(sequence: Sequence, num_spaces: int = 2, max_length: int = 5) -> str:
    r"""Creates a concise summary of a sequence.

    Args:
        sequence: Specifies the sequence to summarize.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.
        max_length (int, optional): Specifies the number of values
            to show in the summary. This option is useful when the
            sequences are long because we may do not want to show
            all the values in the summary. Note this input is also
            used to create the summary of all the values in the
            sequence. Default: ``5``

    Returns:
        str: The concise summary of the sequence.
    """
    summary = to_torch_sequence_str(
        [
            concise_summary(val, num_spaces=num_spaces, max_length=max_length)
            for val in sequence[:max_length]
        ]
    )
    suffix = "\n..." if len(sequence) > max_length else ""
    return str_add_indent(
        f"{type(sequence)} | length={len(sequence):,}\n{summary}{suffix}", num_spaces=num_spaces
    )


def _concise_summary_mapping(mapping: Mapping, num_spaces: int = 2, max_length: int = 5) -> str:
    r"""Creates a concise summary of a mapping.

    Args:
        mapping: Specifies the mapping to summarize.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.
        max_length (int, optional): Specifies the number of values to
            show in the summary. This option is useful when the
            mapping are too big because we may do not want to show
            all the values in the summary. You can set this input to
            ``-1`` to show all the values in the summary. Note this
            input is also used to create the summary of all the values
            in the sequence. Default: ``5``

    Returns:
        str: The concise summary of the mapping.
    """
    summary = to_torch_mapping_str(
        {
            key: concise_summary(value, num_spaces=num_spaces, max_length=max_length)
            for key, value in islice(mapping.items(), max_length)
        }
    )
    suffix = "\n..." if len(mapping) > max_length else ""
    return str_add_indent(
        f"{type(mapping)} | length={len(mapping):,}\n{summary}{suffix}", num_spaces=num_spaces
    )



from __future__ import annotations

__all__ = ["BaseIterDataPipeCreator", "setup_iter_datapipe_creator"]

import logging
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from torch.utils.data import IterDataPipe

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class BaseIterDataPipeCreator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to implement an ``IterDataPipe`` creator.

    A ``IterDataPipe`` creator is responsible to create a single
    DataPipe.

    Note: it is possible to create an ``IterDataPipe`` object without
    using this class.
    """

    @abstractmethod
    def create(
        self, engine: BaseEngine | None = None, source_inputs: Sequence | None = None
    ) -> IterDataPipe:
        r"""Creates an ``IterDataPipe`` object.

        Args:
            engine (``BaseEngine`` or ``None``, optional): Specifies
                an engine. The engine can be used to initialize the
                ``IterDataPipe`` by using the current epoch value.
                Default: ``None``
            source_inputs (sequence or ``None``): Specifies the first
                positional arguments of the source ``IterDataPipe``.
                This argument can be used to create a new
                ``IterDataPipe`` object, that takes existing
                ``IterDataPipe`` objects as input. See examples below
                to see how to use it. If ``None``, ``source_inputs``
                is set to an empty tuple. Default: ``None``

        Returns:
            ``IterDataPipe``: The created ``IterDataPipe`` object.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.creators.datapipe import BaseIterDataPipeCreator
            >>> creator: BaseIterDataPipeCreator = ...  # Create an IterDataPipe creator
            # Create an IterDataPipe object
            >>> from torch.utils.data import IterDataPipe
            >>> my_datapipe: IterDataPipe = creator.create()
            # Create an IterDataPipe object with an engine
            >>> from gravitorch.engines import BaseEngine
            >>> my_engine: BaseEngine = ...  # Create an engine
            >>> my_datapipe: IterDataPipe = creator.create(my_engine)
            # Create an IterDataPipe object with some source_inputs
            >>> my_datapipe: IterDataPipe = creator.create(source_inputs=[...])
            # Create an IterDataPipe object with en engine and some source_inputs
            >>> my_datapipe: IterDataPipe = creator.create(engine=my_engine, source_inputs=[...])
        """


def setup_iter_datapipe_creator(creator: BaseIterDataPipeCreator | dict) -> BaseIterDataPipeCreator:
    r"""Sets up an ``IterDataPipe`` creator.

    The ``IterDataPipe`` creator is instantiated from its
    configuration by using the ``BaseDataPipeCreator`` factory
    function.

    Args:
        creator (``BaseIterDataPipeCreator`` or dict): Specifies the
            ``IterDataPipe`` creator or its configuration.

    Returns:
        ``BaseIterDataPipeCreator``: The instantiated ``IterDataPipe``
            creator.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.creators.datapipe import setup_iter_datapipe_creator
        # Set up an ``IterDataPipe`` creator from an ``IterDataPipe`` creator i.e. do nothing ;)
        >>> from gravitorch.creators.datapipe import SequentialIterDataPipeCreator
        >>> creator = setup_iter_datapipe_creator(
        ...     SequentialIterDataPipeCreator(
        ...         config=[
        ...             {
        ...                 OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
        ...                 "data": [1, 2, 3, 4],
        ...             }
        ...         ],
        ...     )
        ... )
        # Set up an ``IterDataPipe`` creator from its config
        >>> from objectory import OBJECT_TARGET
        >>> creator = setup_iter_datapipe_creator(
        ...     {
        ...         OBJECT_TARGET: "gravitorch.creators.datapipe.SequentialIterDataPipeCreator",
        ...         "config": [
        ...             {
        ...                 OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
        ...                 "data": [1, 2, 3, 4],
        ...             }
        ...         ],
        ...     }
        ... )
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing a IterDataPipe creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseIterDataPipeCreator.factory(**creator)
    return creator



r"""This module defines some steps that are used in the experiment tracker."""

__all__ = ["Step", "EpochStep", "IterationStep"]

from dataclasses import dataclass


@dataclass
class Step:
    r"""Defines a generic step.

    A step should have a step number and a name.

    Args:
        step (int): Specifies the step number.
        name (str): Specifies the name.
    """
    step: int
    name: str


class EpochStep(Step):
    r"""Defines the epoch step.

    Args:
        step (int): Specifies the epoch number.
    """

    def __init__(self, epoch: int):
        super().__init__(step=epoch, name="epoch")


class IterationStep(Step):
    r"""Defines the iteration step.

    Args:
        step (int): Specifies the iteration number.
    """

    def __init__(self, iteration: int):
        super().__init__(step=iteration, name="iteration")



__all__ = [
    "AverageMeter",
    "BinaryConfusionMatrix",
    "EmptyMeterError",
    "ExponentialMovingAverage",
    "ExtremaTensorMeter",
    "MeanTensorMeter",
    "MovingAverage",
    "MulticlassConfusionMatrix",
    "ScalarMeter",
    "TensorMeter",
    "TensorMeter2",
]

from gravitorch.utils.meters.average import AverageMeter
from gravitorch.utils.meters.confusion_matrix import (
    BinaryConfusionMatrix,
    MulticlassConfusionMatrix,
)
from gravitorch.utils.meters.exceptions import EmptyMeterError
from gravitorch.utils.meters.moving_average import (
    ExponentialMovingAverage,
    MovingAverage,
)
from gravitorch.utils.meters.scalar import ScalarMeter
from gravitorch.utils.meters.tensor import (
    ExtremaTensorMeter,
    MeanTensorMeter,
    TensorMeter,
    TensorMeter2,
)



__all__ = ["TrivialPartitioner"]

from collections.abc import Sequence
from typing import Optional, TypeVar

from gravitorch.data.partitioners.base import BasePartitioner
from gravitorch.engines import BaseEngine

T = TypeVar("T")


class TrivialPartitioner(BasePartitioner[T]):
    r"""Implements a partitioner that creates the trivial partition."""

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def partition(
        self, items: Sequence[T], engine: Optional[BaseEngine] = None
    ) -> list[Sequence[T]]:
        return [items]



r"""This module contains handlers to update a learning rate (LR) scheduler."""

__all__ = [
    "EpochLRSchedulerUpdater",
    "MetricEpochLRSchedulerUpdater",
    "IterationLRSchedulerUpdater",
    "LRSchedulerUpdater",
    "MetricLRSchedulerUpdater",
]

import logging

from gravitorch import constants as ct
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.events import VanillaEventHandler

logger = logging.getLogger(__name__)


class LRSchedulerUpdater(BaseHandler):
    r"""Implements a handler to update the learning rate (LR) scheduler at a
    given event.

    Args:
        event (str): Specifies the event when the learning rate
            scheduler is updated.
    """

    def __init__(self, event: str):
        self._event = str(event)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        if engine.lr_scheduler is None:
            logger.info("No event is added to the engine because the LR scheduler is not defined")
            return
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(engine.lr_scheduler.step),
        )


class EpochLRSchedulerUpdater(LRSchedulerUpdater):
    r"""Implements a handler to update the learning rate (LR) scheduler at the
    end of each training epoch."""

    def __init__(self):
        super().__init__(event=EngineEvents.TRAIN_EPOCH_COMPLETED)


class IterationLRSchedulerUpdater(LRSchedulerUpdater):
    r"""Implements a handler to update the learning rate (LR) scheduler at the
    end of each training iteration."""

    def __init__(self):
        super().__init__(event=EngineEvents.TRAIN_ITERATION_COMPLETED)


class MetricLRSchedulerUpdater(BaseHandler):
    r"""Implements a handler to update the learning rate (LR) scheduler with a
    metric value at a given event.

    This updater was designed to be used with ``ReduceLROnPlateau``
    or any LR schedulers that use a metric value as input of the
    ``step`` method.

    Args:
        event (str): Specifies the event when the learning rate
            scheduler is updated. The event should happen after the
            metric was computed.
        metric_name (str, optional): Specifies the metric to use to
            control the LR scheduler. This metric should be accessible
            with the ``get_history`` method of the engine.
            Default: ``'eval/loss'``
    """

    def __init__(self, event: str, metric_name: str = f"{ct.EVAL}/loss"):
        self._event = event
        self._metric_name = metric_name

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, metric_name={self._metric_name})"
        )

    def attach(self, engine: BaseEngine) -> None:
        if engine.lr_scheduler is None:
            logger.info("No event is added to the engine because the LR scheduler is not defined")
            return
        # Note that it is not possible to use the LR scheduler `step` method because the `step`
        # method needs a metric value as input.
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(self.step, handler_kwargs={"engine": engine}),
        )

    def step(self, engine: BaseEngine) -> None:
        r"""Updates the LR scheduler with a metric value.

        Args:
            engine (``BaseEngine``): Specifies the engine with the LR
                scheduler to update and the metric used to update the
                LR scheduler.
        """
        engine.lr_scheduler.step(engine.get_history(self._metric_name).get_last_value())


class MetricEpochLRSchedulerUpdater(MetricLRSchedulerUpdater):
    r"""Implements a handler to update the learning rate (LR) scheduler with a
    metric value at the end of each epoch.

    This updater was designed to be used with ``ReduceLROnPlateau`` or
    any LR schedulers that use a metric value as input of the ``step``
    method. The ``step`` method of the LR scheduler is called at the
    ``"epoch_completed"`` event, so you need to use a metric that is
    updated before this event.

    Args:
        metric_name (str, optional): Specifies the metric to use to
            control the LR scheduler. This metric should be accessible
            with the ``get_history`` method of the engine.
            Default: ``'eval/loss'``
    """

    def __init__(self, metric_name: str = f"{ct.EVAL}/loss"):
        super().__init__(event=EngineEvents.EPOCH_COMPLETED, metric_name=metric_name)



__all__ = [
    "BaseDataSource",
    "DataCreatorIterDataPipeCreatorDataSource",
    "DatasetDataSource",
    "ImageNetDataSource",
    "IterDataPipeCreatorDataSource",
    "LoaderNotFoundError",
    "MnistDataSource",
    "setup_and_attach_data_source",
    "setup_data_source",
]

from gravitorch.datasources.base import BaseDataSource, LoaderNotFoundError
from gravitorch.datasources.datapipe import (
    DataCreatorIterDataPipeCreatorDataSource,
    IterDataPipeCreatorDataSource,
)
from gravitorch.datasources.dataset import DatasetDataSource
from gravitorch.datasources.imagenet import ImageNetDataSource
from gravitorch.datasources.mnist import MnistDataSource
from gravitorch.datasources.utils import setup_and_attach_data_source, setup_data_source



__all__ = ["BaseResourceRunner"]

import logging
from abc import abstractmethod
from collections.abc import Sequence
from contextlib import ExitStack
from typing import Any, Optional, Union

from gravitorch.rsrc.base import BaseResource, setup_resource
from gravitorch.runners.base import BaseRunner

logger = logging.getLogger(__name__)


class BaseResourceRunner(BaseRunner):
    r"""Implements a base class to easily implement a runner using resource
    managers.

    Args:
        resources (sequence or ``None``, optional): Specifies a
            sequence of resources or their configurations.
            Default: ``None``
    """

    def __init__(self, resources: Optional[Sequence[Union[BaseResource, dict]]] = None):
        if resources is None:
            resources = tuple()
        self._resources = tuple(setup_resource(resource) for resource in resources)

    def run(self) -> Any:
        r"""Sets up the distributed context and executes the logic of the
        runner.

        Returns:
            Any artifact of the runner
        """
        with ExitStack() as stack:
            for resource in self._resources:
                stack.enter_context(resource)
            output = self._run()
        return output

    @abstractmethod
    def _run(self) -> Any:
        r"""Executes the logic of the runner after the .

        Returns:
            Any artifact of the runner
        """



r"""This module implements an evaluation loop using the Accelerate library
(https://huggingface.co/docs/accelerate)."""

__all__ = ["AccelerateEvaluationLoop"]

import logging
import sys
from collections.abc import Iterable
from typing import Any, Union

import torch
from torch.nn import Module
from tqdm import tqdm

from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.utils.evaluation_loops.base import BaseBasicEvaluationLoop
from gravitorch.utils.evaluation_loops.conditions import BaseEvalCondition
from gravitorch.utils.integrations import check_accelerate, is_accelerate_available
from gravitorch.utils.loop_observers import BaseLoopObserver
from gravitorch.utils.profilers import BaseProfiler

if is_accelerate_available():
    from accelerate import Accelerator
else:
    Accelerator = None  # pragma: no cover

logger = logging.getLogger(__name__)


class AccelerateEvaluationLoop(BaseBasicEvaluationLoop):
    r"""Implements an evaluation loop that uses ``accelerate.Accelerator`` to
    evaluate a model.

    Args:
        accelerator (``accelerate.Accelerate`` or dict or None,
            optional): Specifies the ``accelerate.Accelerate`` object
            or the parameters to instantiate it. Please read the
            ``accelerate.Accelerator`` documentation to know the
            parameters https://huggingface.co/docs/accelerate/accelerator.html.
            If ``None``, it will use the default parameters.
            Default: ``None``
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"eval"``
        grad_enabled (bool, optional): Specifies if the gradient is
            computed or not in the evaluation loop. By default, the
            gradient is not computed to reduce the memory footprint.
            Default: ``False``
        condition (``BaseEvalCondition`` or dict or None): Specifies
            the condition to evaluate the loop or its configuration.
            If ``None``, the ``EveryEpochEvalCondition(every=1)`` is
            used.  Default ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional):
            Specifies the profiler or its configuration. If ``None``,
            the ``NoOpProfiler`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        accelerator: Union[Accelerator, dict, None] = None,
        tag: str = "eval",
        grad_enabled: bool = False,
        condition: Union[BaseEvalCondition, dict, None] = None,
        observer: Union[BaseLoopObserver, dict, None] = None,
        profiler: Union[BaseProfiler, dict, None] = None,
    ):
        check_accelerate()
        self._accelerator = self._setup_accelerator(accelerator or {})
        logger.info(f"accelerator state:\n{self._accelerator.state}")
        super().__init__(tag=tag, condition=condition, observer=observer, profiler=profiler)
        self._grad_enabled = grad_enabled

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  tag={self._tag},\n"
            f"  grad_enabled={self._grad_enabled},\n"
            f"  condition={self._condition},\n"
            f"  observer={self._observer},\n"
            f"  profiler={self._profiler},\n"
            ")"
        )

    def _eval_one_batch(self, engine: BaseEngine, model: Module, batch: Any) -> dict:
        engine.fire_event(EngineEvents.EVAL_ITERATION_STARTED)
        with torch.set_grad_enabled(self._grad_enabled):
            output = model(batch)
        engine.fire_event(EngineEvents.EVAL_ITERATION_COMPLETED)
        return output

    def _prepare_model_data_loader(self, engine: BaseEngine) -> tuple[Module, Iterable]:
        logger.info("Preparing the model and data loader...")
        model, data_loader = self._accelerator.prepare(
            [
                engine.model,
                engine.data_source.get_data_loader(loader_id=self._tag, engine=engine),
            ]
        )

        prefix = f"({dist.get_rank()}/{dist.get_world_size()}) " if dist.is_distributed() else ""
        data_loader = tqdm(
            data_loader,
            desc=f"{prefix}Evaluation [{engine.epoch}]",
            position=dist.get_rank(),
            file=sys.stdout,
        )
        logger.info("Evaluation data loader has been created")
        return model, data_loader

    def _setup_accelerator(self, accelerator: Union[Accelerator, dict]) -> Accelerator:
        r"""Sets up the accelerator.

        Args:
            accelerator (``accelerate.Accelerator`` or dict, optional):
                Specifies the ``accelerate.Accelerator`` object or the
                parameters to instantiate it. Please read the
                ``accelerate.Accelerator`` documentation to know the
                parameters https://huggingface.co/docs/accelerate/accelerator.html.

        Returns:
            ``accelerate.Accelerator``: The accelerator object.

        Raises:
            RuntimeError: if the accelerate package is not installed.
        """
        if isinstance(accelerator, Accelerator):
            return accelerator
        logger.info(f"accelerator options: {accelerator}")
        return Accelerator(**accelerator)



r"""This module defines the squared logarithmic error (SLE) metric."""

__all__ = ["RootMeanSquaredError", "SquaredError"]

import logging
from typing import Union

from torch import Tensor
from torch.nn.functional import mse_loss

from gravitorch.models.metrics.base_epoch import BaseStateEpochMetric
from gravitorch.models.metrics.state import BaseState, ErrorState, RootMeanErrorState

logger = logging.getLogger(__name__)


class SquaredError(BaseStateEpochMetric):
    r"""Implements the squared error metric.

    Args:
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'sq_err'``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import SquaredError
        >>> metric = SquaredError('eval')
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/sq_err_mean': 0.0,
         'eval/sq_err_max': 0.0,
         'eval/sq_err_min': 0.0,
         'eval/sq_err_sum': 0.0,
         'eval/sq_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sq_err_mean': 0.16666666666666666,
         'eval/sq_err_max': 1.0,
         'eval/sq_err_min': 0.0,
         'eval/sq_err_sum': 2.0,
         'eval/sq_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/sq_err_mean': 0.5,
         'eval/sq_err_max': 1.0,
         'eval/sq_err_min': 0.0,
         'eval/sq_err_sum': 2.0,
         'eval/sq_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "sq_err",
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or ErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the squared logarithmic error metric given a mini-batch of
        examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.
        """
        self._state.update(mse_loss(prediction.float(), target.float(), reduction="none"))


class RootMeanSquaredError(SquaredError):
    r"""Implements the squared error metric.

    Args:
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'sq_err'``
        track_num_predictions (bool, optional): If ``True``, the state
            tracks and returns the number of predictions.
            Default: ``True``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics import RootMeanSquaredError
        >>> metric = RootMeanSquaredError('eval')
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/rmse_root_mean': 0.0, 'eval/rmse_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/rmse_root_mean': 0.408248290463863, 'eval/rmse_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()
        {'eval/rmse_root_mean': 0.7071067811865476, 'eval/rmse_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "rmse",
        track_num_predictions: bool = True,
    ):
        super().__init__(
            mode=mode,
            name=name,
            state=RootMeanErrorState(track_num_predictions=track_num_predictions),
        )



__all__ = ["to_tuple"]

from typing import Any


def to_tuple(value: Any) -> tuple:
    r"""Converts a value to a tuple.

    This function is a no-op if the input is a tuple.

    Args:
        value: Specifies the value to convert.

    Returns:
        tuple: The input value in a tuple.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils import to_tuple
    """
    if isinstance(value, tuple):
        return value
    if isinstance(value, (bool, int, float, str)):
        return tuple([value])
    return tuple(value)



r"""This module contains several focal loss implementations."""

__all__ = ["BinaryFocalLoss", "BinaryFocalLossWithLogits"]

import torch
from torch import Tensor
from torch.nn import BCELoss, BCEWithLogitsLoss, Module

from gravitorch.nn.functional import basic_loss_reduction, check_basic_loss_reduction


class _BinaryFocalLoss(Module):
    r"""Base class to implement the binary Focal Loss with and without logits.

    Based on "Focal Loss for Dense Object Detection"
    (https://arxiv.org/pdf/1708.02002.pdf)

    Args:
        alpha (float, optional): Specifies the weighting factor in
            ``[0, 1]``.
        gamma (float, optional): Specifies the focusing parameter
            (``>=0``).
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed.
    """

    def __init__(
        self,
        bce: Module,
        alpha: float,
        gamma: float,
        reduction: str,
    ):
        super().__init__()
        self.bce = bce

        if 0 <= alpha <= 1:
            self._alpha = float(alpha)
        else:
            raise ValueError(
                f"Incorrect parameter alpha ({alpha}). The valid range of value is [0, 1]."
            )

        if gamma >= 0:
            self._gamma = float(gamma)
        else:
            raise ValueError(
                f"Incorrect parameter gamma ({gamma}). Gamma has to be positive (>=0)."
            )

        check_basic_loss_reduction(reduction)
        self.reduction = reduction

    def extra_repr(self) -> str:
        return f"alpha={self._alpha}, gamma={self._gamma}, reduction={self.reduction}"

    def forward(self, prediction: Tensor, target: Tensor) -> Tensor:
        r"""Computes the binary Focal Loss.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(batch size, num classes)`` and type float):
                Specifies the predicted probabilities. All elements
                should be between ``0`` and ``1``.
            target (``torch.Tensor`` of shape
                ``(batch size, num classes)`` and type float):
                Specifies the targets. ``1`` (resp. ``0``) means a
                positive (resp. negative) example.

        Returns:
            ``torch.Tensor`` of type float: The loss value(s). The
                shape of the tensor depends on the reduction. If the
                reduction is ``mean`` or ``sum``, the tensor has a
                single scalar value. If the reduction is ``none``,
                the shape of the tensor is the same that the inputs.
        """
        bce_loss = self.bce(prediction, target)
        pt = torch.exp(-bce_loss)
        alpha_t = self._alpha * target + (1 - self._alpha) * (
            1 - target
        )  # alpha for positive samples, else 1-alpha
        focal_loss = alpha_t * (1 - pt) ** self._gamma * bce_loss
        return basic_loss_reduction(focal_loss, self.reduction)


class BinaryFocalLoss(_BinaryFocalLoss):
    r"""Implementation of the binary Focal Loss.

    Based on "Focal Loss for Dense Object Detection"
    (https://arxiv.org/pdf/1708.02002.pdf)

    Args:
        alpha (float, optional): Specifies the weighting factor in
            ``[0, 1]``. Default: ``0.5``
        gamma (float, optional): Specifies the focusing parameter
            (``>=0``). Default: ``2``
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``"mean"``
    """

    def __init__(
        self,
        alpha: float = 0.5,
        gamma: float = 2.0,
        reduction: str = "mean",
    ):
        super().__init__(
            bce=BCELoss(reduction="none"),
            alpha=alpha,
            gamma=gamma,
            reduction=reduction,
        )


class BinaryFocalLossWithLogits(_BinaryFocalLoss):
    r"""Implementation of the binary Focal Loss with logits.

    Based on "Focal Loss for Dense Object Detection"
    (https://arxiv.org/pdf/1708.02002.pdf)

    The predictions do not have to be between ``0`` and ``1`` because
    they will be converted to probability.

    Args:
        alpha (float, optional): Specifies the weighting factor in
            ``[0, 1]``. Default: ``0.5``
        gamma (float, optional): Specifies the focusing parameter
            (``>=0``). Default: ``2``
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``"mean"``
    """

    def __init__(
        self,
        alpha: float = 0.5,
        gamma: float = 2.0,
        reduction: str = "mean",
    ):
        super().__init__(
            bce=BCEWithLogitsLoss(reduction="none"),
            alpha=alpha,
            gamma=gamma,
            reduction=reduction,
        )



__all__ = ["FusionFFN"]

from typing import Union

from torch import Tensor
from torch.nn import Module

from gravitorch.nn.utils import get_module_output_size, setup_nn_module


class FusionFFN(Module):
    r"""Implements a module that fuses representations and then applies a feed-
    forward network (FFN) on the fused representation.

    Args:
        fusion (``torch.nn.Module``, optional): Specifies the fusion
            module or its configuration.
        ffn (``torch.nn.Module``, optional): Specifies the FFN or its
            configuration.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn import ConcatFusion, FusionFFN
        >>> module = FusionFFN(ConcatFusion(), torch.nn.Linear(6, 4))
        >>> module
        FusionFFN(
          (fusion): ConcatFusion(dim=-1)
          (ffn): Linear(in_features=6, out_features=4, bias=True)
        )
        >>> x1 = torch.tensor([[2, 3, 4], [5, 6, 7]], dtype=torch.float, requires_grad=True)
        >>> x2 = torch.tensor([[12, 13, 14], [15, 16, 17]], dtype=torch.float, requires_grad=True)
        >>> out = module(x1, x2)
        >>> out
        tensor([[ 2.1302,  7.8643,  8.3265,  4.7277],
                [ 1.6980, 11.7446,  9.8449,  5.8402]], grad_fn=<AddmmBackward0>)
        >>> out.mean().backward()
    """

    def __init__(self, fusion: Union[Module, dict], ffn: Union[Module, dict]):
        super().__init__()
        self.fusion = setup_nn_module(fusion)
        self.ffn = setup_nn_module(ffn)

    @property
    def output_size(self) -> int:
        r"""int: The output size of the module."""
        return get_module_output_size(self.ffn)

    def forward(self, *inputs: Tensor) -> Tensor:
        r"""Fuses the inputs and then applied a feed-forward network (FFN) on
        the fused representation.

        Args:
            *inputs (sequence of ``torch.Tensor``): Specifies the
                sequence of tensors to fuse. The shape of the tensors
                may depend on the feed-forward network (FFN).

        Returns:
            ``torch.Tensor``: The fused representation.
        """
        return self.ffn(self.fusion(*inputs))



r"""This module implements a no-operation data summary."""

__all__ = ["NoOpDataSummary"]

from typing import TypeVar

from gravitorch.utils.data_summary.base import BaseDataSummary

T = TypeVar("T")


class NoOpDataSummary(BaseDataSummary[T]):
    r"""Implements a no-operation data summary."""

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def add(self, data: T) -> None:
        r"""Adds new data to the summary.

        Args:
            data: Specifies the data to add to the summary.
        """

    def reset(self) -> None:
        r"""Resets the data summary."""

    def summary(self) -> dict:
        return {}



__all__ = ["TextArtifact"]

from pathlib import Path

from gravitorch.utils.artifacts.base import BaseArtifact
from gravitorch.utils.io import save_text


class TextArtifact(BaseArtifact):
    r"""Implements a text artifact.

    Args:
        tag (str): Specifies the artifact tag. The tag is used to
            define the text filename.
        data (str): Specifies the data to save in the text artifact.
    """

    def __init__(self, tag: str, data: str):
        self._tag = str(tag)
        self._data = str(data)

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(tag={self._tag})"

    def create(self, path: Path) -> None:
        save_text(to_save=self._data, path=path.joinpath(f"{self._tag}.txt"))



r"""This module defines some utility functions for the data loader creators."""

__all__ = ["setup_data_loader_creator"]

import logging
from typing import Union

from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.creators.dataloader.pytorch import AutoDataLoaderCreator
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_data_loader_creator(
    creator: Union[BaseDataLoaderCreator, dict, None]
) -> BaseDataLoaderCreator:
    r"""Sets up a data loader creator.

    Args:
        creator (``BaseDataLoaderCreator`` or dict or None):
            Specifies the data loader creator or its configuration.
            If ``None``, a data loader creator will be created
            automatically.

    Returns:
        ``BaseDataLoaderCreator``: The data loader creator.
    """
    if creator is None:
        creator = AutoDataLoaderCreator()
    if isinstance(creator, dict):
        logger.info(
            "Initializing a data loader creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDataLoaderCreator.factory(**creator)
    return creator



__all__ = [
    "BinaryAccuracy",
    "BinaryConfusionMatrix",
    "CategoricalAccuracy",
    "CategoricalConfusionMatrix",
    "CategoricalCrossEntropy",
    "TopKAccuracy",
]

from gravitorch.models.metrics.classification.accuracy import (
    BinaryAccuracy,
    CategoricalAccuracy,
    TopKAccuracy,
)
from gravitorch.models.metrics.classification.confusion_matrix import (
    BinaryConfusionMatrix,
    CategoricalConfusionMatrix,
)
from gravitorch.models.metrics.classification.cross_entropy import (
    CategoricalCrossEntropy,
)



__all__ = ["SourceWrapperIterDataPipe"]

import copy
import logging
from collections.abc import Iterable, Iterator

from torch.utils.data import IterDataPipe

from gravitorch.utils.format import str_add_indent
from gravitorch.utils.summary import concise_summary

logger = logging.getLogger(__name__)


class SourceWrapperIterDataPipe(IterDataPipe):
    r"""Creates a simple source DataPipe from an iterable.

    Based on https://github.com/pytorch/pytorch/blob/3c2199b159b6ec57af3f7ea22d61ace9ce5cf5bc/torch/utils/data/datapipes/iter/utils.py#L8-L50  # noqa: B950

    Args:
        source (``iterable``): Specifies the input iterable.
        deepcopy (bool, optional): If ``True``, the input iterable
            object is deep-copied before to iterate over the data.
            It allows a deterministic behavior when in-place
            operations are performed on the data. Default: ``False``
    """

    def __init__(self, source: Iterable, deepcopy: bool = False):
        self._source = source
        self._deepcopy = bool(deepcopy)

    def __iter__(self) -> Iterator:
        source = self._source
        if self._deepcopy:
            try:
                source = copy.deepcopy(source)
            except TypeError:
                logger.warning(
                    "The input iterable can not be deepcopied, please be aware of in-place "
                    "modification would affect source data."
                )
        yield from source

    def __len__(self) -> int:
        return len(self._source)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  deepcopy: {self._deepcopy},\n"
            f"  source:\n    {str_add_indent(concise_summary(self._source), num_spaces=4)}\n)"
        )



__all__ = ["BaseLoopObserver", "NoOpLoopObserver", "PyTorchBatchSaver", "setup_loop_observer"]

from gravitorch.utils.loop_observers.base import BaseLoopObserver
from gravitorch.utils.loop_observers.batch_saving import PyTorchBatchSaver
from gravitorch.utils.loop_observers.factory import setup_loop_observer
from gravitorch.utils.loop_observers.noop import NoOpLoopObserver



r"""Defines some utilities function to analyze a ``torch.nn.Module``.

Inspired from https://github.com/PyTorchLightning/pytorch-
lightning/blob/master/pytorch_lightning/core/memory.py
"""

__all__ = [
    "ModuleSummary",
    "multiline_format_dtype",
    "multiline_format_size",
    "parse_batch_dtype",
    "parse_batch_shape",
]

from typing import Any, Union

import torch
from torch.nn import Module
from torch.utils.hooks import RemovableHandle

from gravitorch.nn.utils.module_helpers import num_learnable_parameters, num_parameters

PARAMETER_NUM_UNITS = (" ", "K", "M", "B", "T")
UNKNOWN_SIZE = "?"
UNKNOWN_DTYPE = "?"


class ModuleSummary:
    r"""Summary class for a single layer in a ``torch.nn.Module``.

    It collects the following information:

    - Type of the layer (e.g. Linear, BatchNorm1d, ...)
    - Input shape
    - Output shape
    - Input data type
    - Output data type
    - Number of parameters
    - Number of learnable parameters

    The input and output shapes are only known after the example input
    array was passed through the model.

    Example::

        >>> model = torch.nn.Conv2d(3, 8, 3)
        >>> summary = ModuleSummary(model)
        >>> summary.num_parameters
        224
        >>> summary.num_learnable_parameters
        224
        >>> summary.layer_type
        'Conv2d'
        >>> output = model(torch.rand(1, 3, 5, 5))
        >>> summary.in_size
        [1, 3, 5, 5]
        >>> summary.out_size
        [1, 8, 3, 3]
        >>> summary.in_dtype
        torch.float32
        >>> summary.out_dtype
        torch.float32

    Args:
        module: A module to summarize
    """

    def __init__(self, module: Module):
        super().__init__()
        self._module = module
        self._hook_handle = self._register_hook()
        self._in_size = None
        self._out_size = None
        self._in_dtype = None
        self._out_dtype = None

    def __del__(self):
        self.detach_hook()

    def _register_hook(self) -> RemovableHandle:
        r"""Registers a hook on the module that computes the input and output
        size(s) on the first forward pass.

        If the hook is called, it will remove itself from the module,
        meaning that recursive models will only record their input and
        output shapes once.

        Return:
            ``RemovableHandle``: A handle for the installed hook.
        """

        def hook(module, inp, out):
            if len(inp) == 1:
                inp = inp[0]
            self._in_size = parse_batch_shape(inp)
            self._out_size = parse_batch_shape(out)
            self._in_dtype = parse_batch_dtype(inp)
            self._out_dtype = parse_batch_dtype(out)
            self._hook_handle.remove()

        return self._module.register_forward_hook(hook)

    def detach_hook(self) -> None:
        r"""Removes the forward hook if it was not already removed in the
        forward pass.

        Will be called after the summary is created.
        """
        self._hook_handle.remove()

    @property
    def in_dtype(self) -> Union[str, list]:
        return self._in_dtype or UNKNOWN_DTYPE

    @property
    def out_dtype(self) -> Union[str, list]:
        return self._out_dtype or UNKNOWN_DTYPE

    @property
    def in_size(self) -> Union[str, list]:
        return self._in_size or UNKNOWN_SIZE

    @property
    def out_size(self) -> Union[str, list]:
        return self._out_size or UNKNOWN_SIZE

    @property
    def layer_type(self) -> str:
        r"""str: The class name of the module."""
        return str(self._module.__class__.__qualname__)

    @property
    def num_parameters(self) -> int:
        r"""int: The number of parameters in this module."""
        return num_parameters(self._module)

    @property
    def num_learnable_parameters(self) -> int:
        r"""int: The number of learnable parameters in this module."""
        return num_learnable_parameters(self._module)


def parse_batch_shape(batch: Any) -> Union[str, tuple]:
    r"""Parses the shapes in the batch.

    For now, it only parses the shapes of a tensor, list of tensors
    and tuple of tensors.

    Args:
        batch: Specifies the batch to parse.

    Returns:
        str or tuple: The shapes in the batch or ``"?"`` if it cannot
            parse the input..
    """
    if hasattr(batch, "shape"):
        return tuple(batch.shape)
    if isinstance(batch, (list, tuple)):
        return tuple(parse_batch_shape(el) for el in batch)
    return UNKNOWN_SIZE


def parse_batch_dtype(batch: Any) -> Union[str, tuple]:
    r"""Parses the data types in the batch.

    For now, it only parses the data type of tensor, list of tensors
    and tuple of tensors.

    Args:
        batch: Specifies the batch to parse.

    Returns:
        str or tuple: The data types in the batch.
    """
    if torch.is_tensor(batch):
        return str(batch.dtype)
    if isinstance(batch, (list, tuple)):
        return tuple(parse_batch_dtype(el) for el in batch)
    return UNKNOWN_DTYPE


def multiline_format_size(rows: Union[str, tuple, list]) -> tuple[str, ...]:
    formatted_rows = []
    for row in rows:
        if isinstance(row, (list, tuple)) and isinstance(row[0], (list, tuple)):
            formatted_rows.append("\n".join([str(r) for r in row]))
        else:
            formatted_rows.append(str(row))
    return tuple(formatted_rows)


def multiline_format_dtype(rows: Union[str, tuple, list]) -> tuple[str, ...]:
    formatted_rows = []
    for row in rows:
        if isinstance(row, (list, tuple)):
            formatted_rows.append("\n".join([str(r) for r in row]))
        else:
            formatted_rows.append(str(row))
    return tuple(formatted_rows)



from __future__ import annotations

__all__ = ["BaseDataSourceCreator", "setup_data_source_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory

from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.datasources import BaseDataSource
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class BaseDataSourceCreator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create a data source.

    Note that it is not the unique approach to create a data source.
    Feel free to use other approaches if this approach does not fit your
    needs.
    """

    @abstractmethod
    def create(self, engine: BaseEngine) -> BaseDataSource:
        r"""Creates a data source object.

        This method is responsible to register the event handlers
        associated to the data source.

        Args:
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.

        Returns:
            ``gravitorch.datasources.BaseDataSource``: The created data
                source.
        """


def setup_data_source_creator(creator: BaseDataSourceCreator | dict) -> BaseDataSourceCreator:
    r"""Sets up the data source creator.

    The data source creator is instantiated from its configuration by
    using the ``BaseDataSourceCreator`` factory function.

    Args:
        creator (``BaseDataSourceCreator`` or dict): Specifies the
            data source creator or its configuration.

    Returns:
        ``BaseDataSourceCreator``: The instantiated data source
            creator.
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing the data source creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseDataSourceCreator.factory(**creator)
    return creator



r"""This module defines the base class for the data loader creators."""

from __future__ import annotations

__all__ = ["BaseDataLoaderCreator"]

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from torch.utils.data import Dataset

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

T = TypeVar("T")


class BaseDataLoaderCreator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create data loader.

    This class and its child classes are designed to be used in a data
    source.
    """

    @abstractmethod
    def create(self, dataset: Dataset, engine: BaseEngine | None = None) -> Iterable[T]:
        r"""Creates a data loader given a dataset and an engine.

        The engine can be used to get the epoch, or other information
        about the training/evaluation.

        Args:
            dataset (``torch.utils.data.Dataset``): Specifies the
                dataset.
            engine (``gravitorch.engines.BaseEngine`` or ``None``,
                optional): Specifies an engine. Default: ``None``

        Returns:
            ``Iterable``: The instantiated data loader.
        """



__all__ = [
    "Backend",
    "UnknownBackendError",
    "all_gather",
    "all_reduce",
    "auto_backend",
    "auto_ddp_model",
    "available_backends",
    "backend",
    "barrier",
    "broadcast",
    "device",
    "distributed_context",
    "finalize",
    "get_local_rank",
    "get_nnodes",
    "get_node_rank",
    "get_nproc_per_node",
    "get_rank",
    "get_world_size",
    "gloocontext",
    "hostname",
    "initialize",
    "is_distributed",
    "is_main_process",
    "model_name",
    "ncclcontext",
    "resolve_backend",
    "set_local_rank",
    "show_config",
    # Nvidia
    "CUDA_VISIBLE_DEVICES",
    # PyTorch
    "GROUP_RANK",
    "LOCAL_RANK",
    "LOCAL_WORLD_SIZE",
    "MASTER_ADDR",
    "MASTER_PORT",
    "RANK",
    "ROLE_RANK",
    "ROLE_WORLD_SIZE",
    "TORCHELASTIC_MAX_RESTARTS",
    "TORCHELASTIC_RESTART_COUNT",
    "TORCHELASTIC_RUN_ID",
    "TORCH_DISTRIBUTED_ENV_VARS",
    "WORLD_SIZE",
    # SLURM
    "SLURM_DISTRIBUTED_ENV_VARS",
    "SLURM_JOB_ID",
    "SLURM_JOB_NODELIST",
    "SLURM_LOCALID",
    "SLURM_NTASKS",
    "SLURM_PROCID",
]

from gravitorch.distributed._constants import (
    CUDA_VISIBLE_DEVICES,
    GROUP_RANK,
    LOCAL_RANK,
    LOCAL_WORLD_SIZE,
    MASTER_ADDR,
    MASTER_PORT,
    RANK,
    ROLE_RANK,
    ROLE_WORLD_SIZE,
    SLURM_DISTRIBUTED_ENV_VARS,
    SLURM_JOB_ID,
    SLURM_JOB_NODELIST,
    SLURM_LOCALID,
    SLURM_NTASKS,
    SLURM_PROCID,
    TORCH_DISTRIBUTED_ENV_VARS,
    TORCHELASTIC_MAX_RESTARTS,
    TORCHELASTIC_RESTART_COUNT,
    TORCHELASTIC_RUN_ID,
    WORLD_SIZE,
)
from gravitorch.distributed.auto import auto_ddp_model
from gravitorch.distributed.comm import (
    Backend,
    UnknownBackendError,
    all_gather,
    all_reduce,
    auto_backend,
    available_backends,
    backend,
    barrier,
    broadcast,
    device,
    distributed_context,
    finalize,
    get_local_rank,
    get_nnodes,
    get_node_rank,
    get_nproc_per_node,
    get_rank,
    get_world_size,
    gloocontext,
    hostname,
    initialize,
    is_distributed,
    is_main_process,
    model_name,
    ncclcontext,
    resolve_backend,
    set_local_rank,
    show_config,
)
